"""Storage options types for various filesystems based on fsspec."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import TypedDict

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from typing import Any
    from typing import Literal

    from fsspec.implementations.cache_mapper import AbstractCacheMapper
    from fsspec.spec import AbstractFileSystem


__all__ = [
    "SimpleCacheStorageOptions",
    "GCSStorageOptions",
    "S3StorageOptions",
    "AzureStorageOptions",
    "HfStorageOptions",
    "DataStorageOptions",
    "FTPStorageOptions",
    "GitHubStorageOptions",
    "HDFSStorageOptions",
    "HTTPStorageOptions",
    "FileStorageOptions",
    "MemoryStorageOptions",
    "SFTPStorageOptions",
    "SMBStorageOptions",
    "WebdavStorageOptions",
    "ZipStorageOptions",
    "TarStorageOptions",
]


class _AbstractStorageOptions(TypedDict, total=False):
    """Base storage options for fsspec-based filesystems"""

    # dircache related options
    use_listings_cache: bool
    listings_expiry_time: int | float | None
    max_paths: int | None
    # fs instance cache options
    skip_instance_cache: bool
    # async fs instance options
    asynchronous: bool
    loop: AbstractEventLoop | None
    batch_size: int | None


class _ChainableStorageOptions(TypedDict, total=False):
    """Storage options for filesystems supporting chaining"""

    target_protocol: str | None
    target_options: dict[str, Any] | None
    fs: AbstractFileSystem | None


class SimpleCacheStorageOptions(
    _AbstractStorageOptions,
    _ChainableStorageOptions,
    total=False,
):
    """Storage options for SimpleCache"""

    cache_storage: Literal["TMP"] | list[str] | str
    cache_check: int | Literal[False]
    check_files: bool
    expiry_time: int | Literal[False]
    same_names: bool | None
    compression: str  # todo: specify allowed values
    cache_mapper: AbstractCacheMapper | None


class GCSStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for Google Cloud Storage"""

    # Authentication and project settings
    project: str
    access: Literal["read_only", "read_write", "full_control"]
    token: (
        None
        | Literal["google_default", "cache", "anon", "browser", "cloud"]
        | str
        | dict[str, Any]
    )

    # Performance and caching
    block_size: int | None
    consistency: Literal["none", "size", "md5"]
    cache_timeout: float | None  # overrides listings_expiry_time if set

    # Request configuration
    requests_timeout: float | None
    requester_pays: bool | str
    session_kwargs: dict[str, Any] | None  # aiohttp.ClientSession kwargs
    timeout: float | None  # timeout used for .buckets?

    # Connection settings
    endpoint_url: str | None
    check_connection: bool | None

    # Storage configuration
    default_location: str | None
    version_aware: bool

    # Deprecated options
    # secure_serialize: bool


class S3StorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for AWS S3 and S3-compatible services"""

    # Authentication
    anon: bool
    key: str | None
    secret: str | None
    token: str | None
    username: str | None  # alias for key
    password: str | None  # alias for secret

    # Connection settings
    endpoint_url: str | None
    use_ssl: bool

    # AWS-specific configuration
    client_kwargs: dict[str, Any] | None  # botocore Client kwargs
    config_kwargs: dict[str, Any] | None  # botocore Config kwargs
    s3_additional_kwargs: dict[str, Any] | None  # s3 api methods kwargs
    session: Any | None  # aiobotocore AioSession

    # Performance settings
    requester_pays: bool
    default_block_size: int | None
    default_fill_cache: bool
    default_cache_type: str  # fsspec.caching Literal["readahead", "none", "bytes", ...]
    max_concurrency: int
    fixed_upload_size: bool

    # Feature flags
    version_aware: bool
    cache_regions: bool


class AzureStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for Azure Blob Storage and Azure Data Lake Gen2"""

    # Account and authentication
    account_name: str | None
    account_key: str | None
    connection_string: str | None
    credential: (
        str | Any | None
    )  # azure.core.credentials_async.AsyncTokenCredential or SAS token
    sas_token: str | None
    anon: bool | None

    # Service Principal authentication
    client_id: str | None
    client_secret: str | None
    tenant_id: str | None

    # Connection and networking
    # request_session: Any | None  # for http requests not used by anything ???
    # socket_timeout: int | None  deprecated
    account_host: str | None
    location_mode: Literal["primary", "secondary"]

    # Performance settings
    blocksize: int  # block size for download/upload operations
    default_fill_cache: bool
    default_cache_type: str  # fsspec cache type
    max_concurrency: int | None

    # Timeout settings
    timeout: int | None  # server-side timeout for operations
    connection_timeout: int | None  # connection establishment timeout
    read_timeout: int | None  # read operation timeout

    # Feature flags
    version_aware: bool  # support blob versioning
    assume_container_exists: bool | None  # container existence assumptions


class HfStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for Hugging face filesystem"""

    # Authentication
    token: str | None

    # Connection settings
    endpoint: str | None

    # Performance settings
    block_size: (
        int | None
    )  # Block size for reading bytes; 0 = raw requests file-like objects


class DataStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for Data URIs filesystem"""

    # No specific options for Data URIs at the moment


class FTPStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for FTP filesystem"""

    # Connection settings
    host: str  # The remote server name/ip to connect to (required)
    port: int  # Port to connect with (default: 21)

    # Authentication
    username: (
        str | None
    )  # User's identifier for authentication (anonymous if not given)
    password: str | None  # User's password on the server
    acct: str | None  # Account string for authentication (some servers require this)

    # Performance settings
    block_size: int | None  # Read-ahead or write buffer size

    # FTP-specific settings
    tempdir: (
        str | None
    )  # Directory on remote to put temporary files when in a transaction
    timeout: int  # Timeout of the FTP connection in seconds (default: 30)
    encoding: str  # Encoding for dir and filenames in FTP connection (default: "utf-8")

    # Security settings
    tls: bool  # Use FTP-TLS (default: False)


class GitHubStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for GitHub repository filesystem"""

    # Repository identification
    org: str  # GitHub organization or username
    repo: str  # Repository name
    sha: str | None  # Commit SHA, branch, or tag (default: current master)

    # Authentication
    username: str | None  # GitHub username for authenticated access
    token: str | None  # GitHub personal access token

    # Connection settings
    timeout: tuple[int, int] | int | None  # (connect, read) timeouts or single timeout


class HDFSStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for Hadoop Distributed File System (HDFS)"""

    # Connection settings
    host: str  # Hostname, IP or "default" to try to read from Hadoop config
    port: int  # Port to connect on, or default from Hadoop config if 0

    # Authentication
    user: str | None  # Username to connect as
    kerb_ticket: str | None  # Kerberos ticket for authentication

    # HDFS-specific settings
    replication: int  # Replication factor for write operations (default: 3)
    extra_conf: (
        dict[str, Any] | None
    )  # Additional configuration passed to HadoopFileSystem


class HTTPStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for HTTP(S) filesystem"""

    # Performance settings
    block_size: (
        int | None
    )  # Block size for reading bytes; 0 = raw requests file-like objects

    # Link parsing behavior
    simple_links: (
        bool  # Consider both HTML <a> tags and URL-like strings vs HTML tags only
    )
    same_scheme: (
        bool  # Only consider paths with matching http/https scheme during ls/glob
    )

    # Caching configuration
    cache_type: str  # Default cache type used in open() (e.g., "bytes")
    cache_options: dict[str, Any] | None  # Default cache options used in open()

    # HTTP client configuration
    client_kwargs: dict[str, Any] | None  # Passed to aiohttp.ClientSession
    get_client: Any | None  # Callable that constructs aiohttp.ClientSession

    # Encoding settings
    encoded: bool  # Whether URLs should be encoded

    # Deprecated options
    # size_policy: Any  # Deprecated parameter


class FileStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for local filesystem (file:// and local:// protocols)"""

    # File system behavior
    auto_mkdir: bool  # Whether to create parent directories when opening files


class MemoryStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for memory filesystem"""

    # No specific options for memory filesystem at the moment


class SFTPStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for SFTP/SSH filesystem"""

    # Connection settings
    host: str  # Hostname or IP address (required)
    port: int | None  # SSH port (default: 22)

    # Authentication
    username: str | None  # Username to authenticate as
    password: (
        str | None
    )  # Password authentication; also used for private key decryption
    passphrase: str | None  # Used for decrypting private keys
    pkey: Any | None  # Private key for authentication (paramiko.PKey)
    key_filename: str | list[str] | None  # Filename(s) of private key(s)/certs to try

    # Connection behavior
    timeout: float | None  # TCP connect timeout in seconds
    allow_agent: bool | None  # Whether to connect to SSH agent (default: True)
    look_for_keys: (
        bool | None
    )  # Whether to search for private keys in ~/.ssh/ (default: True)
    compress: bool | None  # Whether to turn on compression
    sock: Any | None  # Socket or socket-like object for communication

    # GSS-API authentication
    gss_auth: bool | None  # Use GSS-API authentication
    gss_kex: bool | None  # Perform GSS-API Key Exchange and user authentication
    gss_deleg_creds: bool | None  # Delegate GSS-API client credentials
    gss_host: str | None  # Target name in kerberos database (default: hostname)
    gss_trust_dns: bool | None  # Trust DNS to canonicalize hostname (default: True)

    # Timeout settings
    banner_timeout: float | None  # SSH banner timeout in seconds
    auth_timeout: float | None  # Authentication response timeout in seconds
    channel_timeout: float | None  # Channel open response timeout in seconds

    # Advanced configuration
    disabled_algorithms: (
        dict[str, Any] | None
    )  # Algorithms to disable (passed to Transport)
    transport_factory: Any | None  # Callable to generate Transport instance
    auth_strategy: Any | None  # AuthStrategy instance for newer authentication

    # SFTP-specific settings
    temppath: str  # Remote temporary directory path (default: "/tmp")


class SMBStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for SMB/Windows/Samba network shares"""

    # Connection settings
    host: str  # The remote server name/ip to connect to (required)
    port: int | None  # Port to connect with (usually 445, sometimes 139)

    # Authentication
    username: str | None  # Username to connect with (required if not using Kerberos)
    password: str | None  # User's password on the server

    # Connection behavior
    timeout: int  # Connection timeout in seconds (default: 60)
    encrypt: bool | None  # Whether to force encryption

    # File access control
    share_access: Literal["r", "w", "d"] | None  # Default access for file operations
    # None (default): exclusively locks file until closed
    # 'r': Allow other handles with read access
    # 'w': Allow other handles with write access
    # 'd': Allow other handles with delete access

    # Session retry configuration
    register_session_retries: int  # Number of session registration retries (default: 4)
    register_session_retry_wait: (
        int  # Wait time between retries in seconds (default: 1)
    )
    register_session_retry_factor: int  # Exponential backoff factor (default: 10)

    # File system behavior
    auto_mkdir: bool  # Whether to create parent directories when opening files


class WebdavStorageOptions(_AbstractStorageOptions, total=False):
    """Storage options for WebDAV filesystem"""

    # Connection settings
    base_url: str  # Base URL of the WebDAV server (required)

    # Authentication
    auth: (
        tuple[str, str] | Any | None
    )  # Authentication (username, password) tuple or custom auth

    # Client configuration
    client: Any | None  # webdav4.client.Client instance


class ZipStorageOptions(
    _AbstractStorageOptions,
    _ChainableStorageOptions,
    total=False,
):
    """Storage options for ZIP archive filesystem"""

    # Archive file settings
    fo: str | Any  # Path to ZIP file or file-like object
    mode: Literal["r", "w", "a"]  # Open mode: read, write, or append

    # ZIP compression settings
    compression: int  # Compression method (e.g., zipfile.ZIP_STORED, ZIP_DEFLATED)
    allowZip64: bool  # Enable ZIP64 extensions for large files
    compresslevel: int | None  # Compression level (None uses default for method)


class TarStorageOptions(
    _AbstractStorageOptions,
    _ChainableStorageOptions,
    total=False,
):
    """Storage options for TAR archive filesystem (read-only)"""

    # Archive file settings
    fo: str | Any  # Path to TAR file or file-like object
    # Compression settings
    compression: (
        str | None
    )  # Compression method: 'gzip', 'bz2', 'xz', or None for auto-detect
    index_store: str | None  # Path to store/load the file index cache



from __future__ import annotations

import os
import pathlib
import shutil
import sys
import warnings
from collections.abc import Iterator
from collections.abc import Sequence
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Literal
from typing import overload
from urllib.parse import SplitResult

from fsspec import AbstractFileSystem

from upath._chain import DEFAULT_CHAIN_PARSER
from upath._chain import Chain
from upath._chain import ChainSegment
from upath._chain import FSSpecChainParser
from upath._protocol import compatible_protocol
from upath._protocol import get_upath_protocol
from upath.core import UnsupportedOperation
from upath.core import UPath
from upath.core import _UPathMixin
from upath.types import UNSET_DEFAULT
from upath.types import JoinablePathLike
from upath.types import PathInfo
from upath.types import ReadablePath
from upath.types import ReadablePathLike
from upath.types import StatResultType
from upath.types import SupportsPathLike
from upath.types import WritablePath
from upath.types import WritablePathLike

if TYPE_CHECKING:
    from typing import IO
    from typing import BinaryIO
    from typing import TextIO
    from typing import TypeVar

    if sys.version_info >= (3, 11):
        from typing import Self
        from typing import Unpack
    else:
        from typing_extensions import Self
        from typing_extensions import Unpack

    from pydantic import GetCoreSchemaHandler
    from pydantic_core.core_schema import CoreSchema

    from upath.types.storage_options import FileStorageOptions

    _WT = TypeVar("_WT", bound="WritablePath")

__all__ = [
    "LocalPath",
    "PosixUPath",
    "WindowsUPath",
    "FilePath",
]


_LISTDIR_WORKS_ON_FILES: bool | None = None


def _check_listdir_works_on_files() -> bool:
    global _LISTDIR_WORKS_ON_FILES
    from fsspec.implementations.local import LocalFileSystem

    fs = LocalFileSystem()
    try:
        fs.ls(__file__)
    except NotADirectoryError:
        _LISTDIR_WORKS_ON_FILES = w = False
    else:
        _LISTDIR_WORKS_ON_FILES = w = True
    return w


def _warn_protocol_storage_options(
    cls: type,
    protocol: str | None,
    storage_options: dict[str, Any],
) -> None:
    if protocol in {"", None} and not storage_options:
        return
    warnings.warn(
        f"{cls.__name__} on python <= (3, 11) ignores protocol and storage_options",
        UserWarning,
        stacklevel=3,
    )


class _LocalPathInfo(PathInfo):
    """Backported PathInfo implementation for LocalPath.
    todo: currently not handling symlinks correctly.
    """

    def __init__(self, path: LocalPath) -> None:
        self._path = path.path

    def exists(self, *, follow_symlinks: bool = True) -> bool:
        return os.path.exists(self._path)

    def is_dir(self, *, follow_symlinks: bool = True) -> bool:
        return os.path.isdir(self._path)

    def is_file(self, *, follow_symlinks: bool = True) -> bool:
        return os.path.isfile(self._path)

    def is_symlink(self) -> bool:
        return os.path.islink(self._path)


class LocalPath(_UPathMixin, pathlib.Path):
    __slots__ = (
        "_chain",
        "_chain_parser",
        "_fs_cached",
        "_relative_base",
    )
    if TYPE_CHECKING:
        _chain: Chain
        _chain_parser: FSSpecChainParser
        _fs_cached: AbstractFileSystem
        _relative_base: str | None

    parser = os.path  # type: ignore[misc,assignment]

    @property
    def _raw_urlpaths(self) -> Sequence[JoinablePathLike]:
        return self.parts

    @_raw_urlpaths.setter
    def _raw_urlpaths(self, value: Sequence[JoinablePathLike]) -> None:
        pass

    if sys.version_info >= (3, 14):

        def rename(
            self,
            target: WritablePathLike,
        ) -> Self:
            t = super().rename(target)  # type: ignore[arg-type]
            if not isinstance(target, type(self)):
                return self.with_segments(t)
            else:
                return t

    if sys.version_info >= (3, 12):

        def __init__(
            self,
            *args,
            protocol: str | None = None,
            chain_parser: FSSpecChainParser = DEFAULT_CHAIN_PARSER,
            **storage_options: Any,
        ) -> None:
            super(_UPathMixin, self).__init__(*args)
            self._chain = Chain(ChainSegment(str(self), "", storage_options))
            self._chain_parser = chain_parser

    elif sys.version_info >= (3, 10):

        def __init__(
            self,
            *args,
            protocol: str | None = None,
            chain_parser: FSSpecChainParser = DEFAULT_CHAIN_PARSER,
            **storage_options: Any,
        ) -> None:
            # super(_UPathMixin, self).__init__(*args)
            _warn_protocol_storage_options(type(self), protocol, storage_options)
            self._drv, self._root, self._parts = self._parse_args(args)  # type: ignore[attr-defined] # noqa: E501
            self._chain = Chain(ChainSegment(str(self), "", {}))
            self._chain_parser = chain_parser

        @classmethod
        def _from_parts(cls, args):
            obj = super()._from_parts(args)
            obj._chain = Chain(ChainSegment(str(obj), "", {}))
            return obj

        @classmethod
        def _from_parsed_parts(cls, drv, root, parts):
            obj = super()._from_parsed_parts(drv, root, parts)
            obj._chain = Chain(ChainSegment(str(obj), "", {}))
            return obj

    else:

        def __init__(
            self,
            *args,
            protocol: str | None = None,
            chain_parser: FSSpecChainParser = DEFAULT_CHAIN_PARSER,
            **storage_options: Any,
        ) -> None:
            _warn_protocol_storage_options(type(self), protocol, storage_options)
            self._drv, self._root, self._parts = self._parse_args(args)  # type: ignore[attr-defined] # noqa: E501
            self._init()
            self._chain_parser = chain_parser

        def _init(self, **kwargs: Any) -> None:
            super()._init(**kwargs)  # type: ignore[misc]
            self._chain = Chain(ChainSegment(str(self), "", {}))

    def __vfspath__(self) -> str:
        return self.__fspath__()

    def __open_reader__(self) -> BinaryIO:
        return self.open("rb")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UPath):
            return NotImplemented
        eq_path = super().__eq__(other)
        if eq_path is NotImplemented:
            return NotImplemented
        return (
            eq_path
            and self.protocol == other.protocol
            and self.storage_options == other.storage_options
        )

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, UPath):
            return NotImplemented
        ne_path = super().__ne__(other)
        if ne_path is NotImplemented:
            return NotImplemented
        return (
            ne_path
            or self.protocol != other.protocol
            or self.storage_options != other.storage_options
        )

    def __hash__(self) -> int:
        return super().__hash__()

    if sys.version_info >= (3, 14):

        def __open_rb__(self, buffering: int = UNSET_DEFAULT) -> BinaryIO:
            return self.open("rb", buffering=buffering)

    def __open_writer__(self, mode: Literal["a", "w", "x"]) -> BinaryIO:
        if mode == "w":
            return self.open(mode="wb")
        elif mode == "a":
            return self.open(mode="ab")
        elif mode == "x":
            return self.open(mode="xb")
        else:
            raise ValueError(f"invalid mode: {mode}")

    def with_segments(self, *pathsegments: str | os.PathLike[str]) -> Self:
        return type(self)(
            *pathsegments,
            protocol=self._protocol,
            **self._storage_options,
        )

    @property
    def path(self) -> str:
        return self.as_posix()

    @property
    def _url(self) -> SplitResult:
        return SplitResult._make((self.protocol, "", self.path, "", ""))

    def joinpath(self, *other) -> Self:
        if not compatible_protocol("", *other):
            raise ValueError("can't combine incompatible UPath protocols")
        return super().joinpath(
            *(
                str(o) if isinstance(o, UPath) and not o.is_absolute() else o
                for o in other
            )
        )

    def __truediv__(self, other) -> Self:
        if not compatible_protocol("", other):
            raise ValueError("can't combine incompatible UPath protocols")
        return super().__truediv__(
            str(other)
            if isinstance(other, UPath) and not other.is_absolute()
            else other
        )

    def __rtruediv__(self, other) -> Self:
        if not compatible_protocol("", other):
            raise ValueError("can't combine incompatible UPath protocols")
        return super().__rtruediv__(
            str(other)
            if isinstance(other, UPath) and not other.is_absolute()
            else other
        )

    @overload  # type: ignore[override]
    def open(
        self,
        mode: Literal["r", "w", "a"] = "r",
        buffering: int = ...,
        encoding: str = ...,
        errors: str = ...,
        newline: str = ...,
        **fsspec_kwargs: Any,
    ) -> TextIO: ...

    @overload
    def open(
        self,
        mode: Literal["rb", "wb", "ab", "xb"],
        buffering: int = ...,
        encoding: str = ...,
        errors: str = ...,
        newline: str = ...,
        **fsspec_kwargs: Any,
    ) -> BinaryIO: ...

    @overload
    def open(
        self,
        mode: str,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        **fsspec_kwargs: Any,
    ) -> IO[Any]: ...

    def open(
        self,
        mode: str = "r",
        buffering: int = UNSET_DEFAULT,
        encoding: str | None = UNSET_DEFAULT,
        errors: str | None = UNSET_DEFAULT,
        newline: str | None = UNSET_DEFAULT,
        **fsspec_kwargs: Any,
    ) -> IO[Any]:
        if not fsspec_kwargs:
            kwargs: dict[str, str | int | None] = {}
            if buffering is not UNSET_DEFAULT:
                kwargs["buffering"] = buffering
            if encoding is not UNSET_DEFAULT:
                kwargs["encoding"] = encoding
            if errors is not UNSET_DEFAULT:
                kwargs["errors"] = errors
            if newline is not UNSET_DEFAULT:
                kwargs["newline"] = newline
            return super().open(mode, **kwargs)  # type: ignore  # noqa: E501
        return UPath.open.__get__(self)(
            mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            **fsspec_kwargs,
        )

    def rmdir(self, recursive: bool = UNSET_DEFAULT) -> None:
        if recursive is UNSET_DEFAULT or not recursive:
            return super().rmdir()
        else:
            shutil.rmtree(self)

    # we need to override pathlib.Path._copy_from to support it as a
    # WritablePath._copy_from target with support for on_name_collision
    # Issue: https://github.com/barneygale/pathlib-abc/issues/48
    _copy_from = UPath._copy_from

    if sys.version_info < (3, 14):  # noqa: C901

        @overload
        def copy(self, target: _WT, **kwargs: Any) -> _WT: ...

        @overload
        def copy(self, target: SupportsPathLike | str, **kwargs: Any) -> Self: ...

        def copy(
            self, target: _WT | SupportsPathLike | str, **kwargs: Any
        ) -> _WT | Self:
            # hacky workaround for missing pathlib.Path.copy in python < 3.14
            # todo: revisit
            _copy: Any = ReadablePath.copy.__get__(self)
            if isinstance(target, str):
                proto = get_upath_protocol(target)
                if proto != self.protocol:
                    target = UPath(target)
                else:
                    target = self.with_segments(target)
            elif not isinstance(target, UPath):
                target = UPath(target)
            if target.is_dir():
                raise IsADirectoryError(str(target))
            return _copy(target, **kwargs)

        @overload
        def copy_into(self, target_dir: _WT, **kwargs: Any) -> _WT: ...

        @overload
        def copy_into(
            self, target_dir: SupportsPathLike | str, **kwargs: Any
        ) -> Self: ...

        def copy_into(
            self,
            target_dir: _WT | SupportsPathLike | str,
            **kwargs: Any,
        ) -> _WT | Self:
            # hacky workaround for missing pathlib.Path.copy_into in python < 3.14
            # todo: revisit
            _copy_into: Any = ReadablePath.copy_into.__get__(self)
            if isinstance(target_dir, str):
                proto = get_upath_protocol(target_dir)
                if proto != self.protocol:
                    target_dir = UPath(target_dir)
                else:
                    target_dir = self.with_segments(target_dir)
            elif not isinstance(target_dir, UPath):
                target_dir = UPath(target_dir)
            return _copy_into(target_dir, **kwargs)

        @overload
        def move(self, target: _WT, **kwargs: Any) -> _WT: ...

        @overload
        def move(self, target: SupportsPathLike | str, **kwargs: Any) -> Self: ...

        def move(
            self, target: _WT | SupportsPathLike | str, **kwargs: Any
        ) -> _WT | Self:
            target = self.copy(target, **kwargs)
            self.fs.rm(self.path, recursive=self.is_dir())
            return target

        @overload
        def move_into(self, target_dir: _WT, **kwargs: Any) -> _WT: ...

        @overload
        def move_into(
            self, target_dir: SupportsPathLike | str, **kwargs: Any
        ) -> Self: ...

        def move_into(
            self, target_dir: _WT | SupportsPathLike | str, **kwargs: Any
        ) -> _WT | Self:
            name = self.name
            if not name:
                raise ValueError(f"{self!r} has an empty name")
            elif hasattr(target_dir, "with_segments"):
                target = target_dir.with_segments(str(target_dir), name)  # type: ignore
            elif isinstance(target_dir, pathlib.PurePath):
                target = UPath(target_dir, name)
            else:
                target = self.with_segments(str(target_dir), name)
            td = target.parent
            if not td.exists():
                raise FileNotFoundError(str(td))
            elif not td.is_dir():
                raise NotADirectoryError(str(td))
            return self.move(target)

        @property
        def info(self) -> PathInfo:
            return _LocalPathInfo(self)

    if sys.version_info < (3, 13):  # noqa: C901

        def full_match(
            self,
            pattern: str | os.PathLike[str],
            *,
            case_sensitive: bool | None = None,
        ) -> bool:
            # hacky workaround for missing pathlib.Path.full_match in python < 3.13
            # todo: revisit
            return self.match(
                pattern,  # type: ignore[arg-type]
                case_sensitive=case_sensitive,
            )

        @classmethod
        def from_uri(cls, uri: str, **storage_options: Any) -> Self:
            return UPath(uri, **storage_options)  # type: ignore[return-value]

        def is_dir(self, *, follow_symlinks: bool = True) -> bool:
            if not follow_symlinks:
                warnings.warn(
                    f"{type(self).__name__}.is_dir(): follow_symlinks=False"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().is_dir()

        def is_file(self, *, follow_symlinks: bool = True) -> bool:
            if not follow_symlinks:
                warnings.warn(
                    f"{type(self).__name__}.is_file(): follow_symlinks=False"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().is_file()

        def read_text(
            self,
            encoding: str | None = None,
            errors: str | None = None,
            newline: str | None = None,
        ) -> str:
            if newline is not None:
                warnings.warn(
                    f"{type(self).__name__}.read_text(): newline"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().read_text(
                encoding=encoding,
                errors=errors,
            )

        def glob(  # type: ignore[override]
            self,
            pattern: str | os.PathLike,
            *,
            case_sensitive: bool | None = None,
            recurse_symlinks: bool = False,
        ) -> Iterator[Self]:
            if isinstance(pattern, str):
                pattern_str = pattern
            else:
                pattern_str = os.fspath(pattern)
            kw: dict[str, Any] = {}
            if case_sensitive is not None:
                if sys.version_info < (3, 12):
                    warnings.warn(
                        f"{type(self).__name__}.glob(): case_sensitive"
                        " is currently ignored.",
                        UserWarning,
                        stacklevel=2,
                    )
                else:
                    kw["case_sensitive"] = case_sensitive
            if recurse_symlinks:
                warnings.warn(
                    f"{type(self).__name__}.glob(): recurse_symlinks=True"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().glob(pattern_str, **kw)

        def rglob(  # type: ignore[override]
            self,
            pattern: str | os.PathLike,
            *,
            case_sensitive: bool | None = None,
            recurse_symlinks: bool = False,
        ) -> Iterator[Self]:
            if isinstance(pattern, str):
                pattern_str = pattern
            else:
                pattern_str = os.fspath(pattern)
            kw: dict[str, Any] = {}
            if case_sensitive is not None:
                if sys.version_info < (3, 12):
                    warnings.warn(
                        f"{type(self).__name__}.rglob(): case_sensitive"
                        " is currently ignored.",
                        UserWarning,
                        stacklevel=2,
                    )
                else:
                    kw["case_sensitive"] = case_sensitive
            if recurse_symlinks:
                warnings.warn(
                    f"{type(self).__name__}.rglob(): recurse_symlinks=True"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().rglob(pattern_str, **kw)

        def owner(self, *, follow_symlinks: bool = True) -> str:
            if not follow_symlinks:
                warnings.warn(
                    f"{type(self).__name__}.owner() follow_symlinks=False"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().owner()

        def group(self, *, follow_symlinks: bool = True) -> str:
            if not follow_symlinks:
                warnings.warn(
                    f"{type(self).__name__}.group() follow_symlinks=False"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().group()

    if sys.version_info < (3, 12):

        def is_junction(self) -> bool:
            return False

        def walk(
            self,
            top_down: bool = True,
            on_error: Callable[[Exception], Any] | None = None,
            follow_symlinks: bool = False,
        ) -> Iterator[tuple[Self, list[str], list[str]]]:
            _walk = ReadablePath.walk.__get__(self)
            return _walk(top_down, on_error, follow_symlinks)

        def match(
            self,
            path_pattern: str | os.PathLike[str],
            *,
            case_sensitive: bool | None = None,
        ) -> bool:
            if isinstance(path_pattern, str):
                pattern_str = path_pattern
            else:
                pattern_str = os.fspath(path_pattern)
            if case_sensitive is not None:
                warnings.warn(
                    f"{type(self).__name__}.match(): case_sensitive"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().match(pattern_str)

        def exists(self, *, follow_symlinks: bool = True) -> bool:
            if not follow_symlinks:
                warnings.warn(
                    f"{type(self).__name__}.exists(): follow_symlinks=False"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().exists()

        def relative_to(  # type: ignore[override]
            self,
            other: str | os.PathLike[str],
            /,
            *_deprecated: str | os.PathLike[str],
            walk_up: bool = False,
        ) -> Self:
            if walk_up:
                warnings.warn(
                    f"{type(self).__name__}.relative_to() walk_up=True"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().relative_to(other, *_deprecated)

    if sys.version_info < (3, 10):

        def hardlink_to(self, target: ReadablePathLike) -> None:
            try:
                os.link(os.fspath(target), os.fspath(self))  # type: ignore[arg-type]
            except AttributeError:
                raise UnsupportedOperation("hardlink operation not supported")

        # let's skip this one as backporting it breaks one pathlib test
        # @property
        # def parents(self) -> Sequence[Self]:
        #     return list(super().parents)

        def stat(  # type: ignore[override]
            self,
            *,
            follow_symlinks: bool = True,
        ) -> StatResultType:
            if not follow_symlinks:
                warnings.warn(
                    f"{type(self).__name__}.stat() follow_symlinks=False"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().stat()  # type: ignore[return-value]

        def write_text(
            self,
            data: str,
            encoding: str | None = None,
            errors: str | None = None,
            newline: str | None = None,
        ) -> int:
            if newline is not None:
                warnings.warn(
                    f"{type(self).__name__}.write_text() newline"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().write_text(
                data,
                encoding=encoding,
                errors=errors,
            )

        def chmod(
            self,
            mode: int,
            *,
            follow_symlinks: bool = True,
        ) -> None:
            if not follow_symlinks:
                warnings.warn(
                    f"{type(self).__name__}.chmod() follow_symlinks=False"
                    " is currently ignored.",
                    UserWarning,
                    stacklevel=2,
                )
            return super().chmod(mode)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        cs = UPath.__get_pydantic_core_schema__.__func__  # type: ignore[attr-defined]
        return cs(cls, source_type, handler)


UPath.register(LocalPath)


# Mypy will ignore the ABC.register call above, so we need to force it to
# think PosixUPath and WindowsUPath are subclasses of UPath.
# This is really not a good pattern, but it's the best we can do without
# either introducing a duck-type protocol for UPath or come up with a
# better solution for the UPath versions of the pathlib.Path subclasses.

if TYPE_CHECKING:

    class WindowsUPath(LocalPath, pathlib.WindowsPath, UPath):  # type: ignore[misc]
        __slots__ = ()

    class PosixUPath(LocalPath, pathlib.PosixPath, UPath):  # type: ignore[misc]
        __slots__ = ()

else:

    class WindowsUPath(LocalPath, pathlib.WindowsPath):
        __slots__ = ()

        if os.name != "nt":

            def __new__(
                cls,
                *args,
                protocol: str | None = None,
                chain_parser: FSSpecChainParser = DEFAULT_CHAIN_PARSER,
                **storage_options: Any,
            ) -> WindowsUPath:
                raise UnsupportedOperation(
                    f"cannot instantiate {cls.__name__} on your system"
                )

    class PosixUPath(LocalPath, pathlib.PosixPath):
        __slots__ = ()

        if os.name == "nt":

            def __new__(
                cls,
                *args,
                protocol: str | None = None,
                chain_parser: FSSpecChainParser = DEFAULT_CHAIN_PARSER,
                **storage_options: Any,
            ) -> PosixUPath:
                raise UnsupportedOperation(
                    f"cannot instantiate {cls.__name__} on your system"
                )


class FilePath(UPath):
    __slots__ = ()

    if TYPE_CHECKING:

        def __init__(
            self,
            *args: JoinablePathLike,
            protocol: Literal["file", "local"] | None = ...,
            chain_parser: FSSpecChainParser = ...,
            **storage_options: Unpack[FileStorageOptions],
        ) -> None: ...

    def __fspath__(self) -> str:
        return self.path

    def iterdir(self) -> Iterator[Self]:
        if _LISTDIR_WORKS_ON_FILES is None:
            _check_listdir_works_on_files()
        elif _LISTDIR_WORKS_ON_FILES and self.is_file():
            raise NotADirectoryError(f"{self}")
        return super().iterdir()

    @property
    def _url(self) -> SplitResult:
        return SplitResult._make((self.protocol, "", self.path, "", ""))

    @classmethod
    def cwd(cls) -> Self:
        return cls(os.getcwd(), protocol="file")

    @classmethod
    def home(cls) -> Self:
        return cls(os.path.expanduser("~"), protocol="file")

    def chmod(
        self,
        mode: int,
        *,
        follow_symlinks: bool = True,
    ) -> None:
        return os.chmod(self.path, mode, follow_symlinks=follow_symlinks)


LocalPath.register(FilePath)



from __future__ import annotations

import sys
from collections.abc import Iterator
from collections.abc import Sequence
from typing import TYPE_CHECKING
from urllib.parse import quote_plus

from upath._protocol import get_upath_protocol
from upath.core import UnsupportedOperation
from upath.core import UPath
from upath.types import JoinablePathLike

if TYPE_CHECKING:
    from typing import Literal

    if sys.version_info >= (3, 11):
        from typing import Self
        from typing import Unpack
    else:
        from typing_extensions import Self
        from typing_extensions import Unpack

    from upath._chain import FSSpecChainParser
    from upath.types.storage_options import DataStorageOptions

__all__ = ["DataPath"]


class DataPath(UPath):
    __slots__ = ()

    if TYPE_CHECKING:

        def __init__(
            self,
            *args: JoinablePathLike,
            protocol: Literal["data"] | None = ...,
            chain_parser: FSSpecChainParser = ...,
            **storage_options: Unpack[DataStorageOptions],
        ) -> None: ...

    @property
    def parts(self) -> Sequence[str]:
        return (self.path,)

    def __str__(self) -> str:
        return self.parser.join(*self._raw_urlpaths)

    def with_segments(self, *pathsegments: JoinablePathLike) -> Self:
        try:
            (segment,) = pathsegments
        except ValueError:
            raise UnsupportedOperation("join not supported by DataPath")
        if get_upath_protocol(segment) != "data":
            raise ValueError(f"requires a data URI, got: {segment!r}")
        return type(self)(segment, protocol="data", **self.storage_options)

    @property
    def name(self) -> str:
        return quote_plus(self.path)

    @property
    def stem(self) -> str:
        return quote_plus(self.path)

    @property
    def suffix(self) -> str:
        return ""

    @property
    def suffixes(self) -> list[str]:
        return []

    def with_name(self, name: str) -> Self:
        raise UnsupportedOperation("with_name not supported by DataPath")

    def with_suffix(self, suffix: str) -> Self:
        raise UnsupportedOperation("with_suffix not supported by DataPath")

    def with_stem(self, stem: str) -> Self:
        raise UnsupportedOperation("with_stem not supported by DataPath")

    @property
    def parent(self) -> Self:
        return self

    @property
    def parents(self) -> Sequence[Self]:
        return []

    def full_match(self, pattern, *, case_sensitive: bool | None = None) -> bool:
        return super().full_match(pattern, case_sensitive=case_sensitive)

    def mkdir(
        self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False
    ) -> None:
        raise FileExistsError(str(self))

    def write_bytes(self, data: bytes) -> int:
        raise UnsupportedOperation("DataPath does not support writing")

    def write_text(
        self,
        data: str,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> int:
        raise UnsupportedOperation("DataPath does not support writing")

    def iterdir(self) -> Iterator[Self]:
        raise NotADirectoryError

    def glob(
        self, pattern, *, case_sensitive=None, recurse_symlinks=False
    ) -> Iterator[Self]:
        return iter([])

    def rglob(
        self, pattern, *, case_sensitive=None, recurse_symlinks=False
    ) -> Iterator[Self]:
        return iter([])

    def unlink(self, missing_ok: bool = False) -> None:
        raise UnsupportedOperation



from __future__ import annotations

import enum
import sys
from collections.abc import Callable
from os import PathLike
from typing import TYPE_CHECKING
from typing import Any
from typing import Optional
from typing import Protocol
from typing import Union
from typing import runtime_checkable

from upath.types._abc import JoinablePath
from upath.types._abc import PathInfo
from upath.types._abc import PathParser
from upath.types._abc import ReadablePath
from upath.types._abc import WritablePath

if TYPE_CHECKING:

    if sys.version_info >= (3, 12):
        from typing import TypeAlias
    else:
        from typing_extensions import TypeAlias

__all__ = [
    "JoinablePath",
    "ReadablePath",
    "WritablePath",
    "JoinablePathLike",
    "ReadablePathLike",
    "WritablePathLike",
    "SupportsPathLike",
    "PathInfo",
    "StatResultType",
    "PathParser",
    "UPathParser",
    "UNSET_DEFAULT",
    "OnNameCollisionFunc",
]


class VFSPathLike(Protocol):
    def __vfspath__(self) -> str: ...


SupportsPathLike: TypeAlias = Union[VFSPathLike, PathLike[str]]
JoinablePathLike: TypeAlias = Union[JoinablePath, SupportsPathLike, str]
ReadablePathLike: TypeAlias = Union[ReadablePath, SupportsPathLike, str]
WritablePathLike: TypeAlias = Union[WritablePath, SupportsPathLike, str]


class _DefaultValue(enum.Enum):
    UNSET = enum.auto()


UNSET_DEFAULT: Any = _DefaultValue.UNSET

# We can't assume this, because pathlib_abc==0.5.1 is ahead of stdlib 3.14
# if sys.version_info >= (3, 14):
#     JoinablePath.register(pathlib.PurePath)
#     ReadablePath.register(pathlib.Path)
#     WritablePath.register(pathlib.Path)


@runtime_checkable
class StatResultType(Protocol):
    """duck-type for os.stat_result"""

    @property
    def st_mode(self) -> int: ...
    @property
    def st_ino(self) -> int: ...
    @property
    def st_dev(self) -> int: ...
    @property
    def st_nlink(self) -> int: ...
    @property
    def st_uid(self) -> int: ...
    @property
    def st_gid(self) -> int: ...
    @property
    def st_size(self) -> int: ...
    @property
    def st_atime(self) -> float: ...
    @property
    def st_mtime(self) -> float: ...
    @property
    def st_ctime(self) -> float: ...
    @property
    def st_atime_ns(self) -> int: ...
    @property
    def st_mtime_ns(self) -> int: ...
    @property
    def st_ctime_ns(self) -> int: ...

    # st_birthtime is available on Windows (3.12+), FreeBSD, and macOS
    # On Linux it's currently unavailable
    # see: https://discuss.python.org/t/st-birthtime-not-available/104350/2
    if (sys.platform == "win32" and sys.version_info >= (3, 12)) or (
        sys.platform == "darwin" or sys.platform.startswith("freebsd")
    ):

        @property
        def st_birthtime(self) -> float: ...


@runtime_checkable
class UPathParser(PathParser, Protocol):
    """duck-type for upath.core.UPathParser"""

    def split(self, path: JoinablePathLike) -> tuple[str, str]: ...
    def splitext(self, path: JoinablePathLike) -> tuple[str, str]: ...
    def normcase(self, path: JoinablePathLike) -> str: ...

    def strip_protocol(self, path: JoinablePathLike) -> str: ...

    def join(
        self,
        path: JoinablePathLike,
        *paths: JoinablePathLike,
    ) -> str: ...

    def isabs(self, path: JoinablePathLike) -> bool: ...

    def splitdrive(self, path: JoinablePathLike) -> tuple[str, str]: ...

    def splitroot(self, path: JoinablePathLike) -> tuple[str, str, str]: ...


OnNameCollisionFunc: TypeAlias = Callable[
    [ReadablePath, WritablePath], tuple[Optional[WritablePath], Optional[WritablePath]]
]



import random
import string

import pytest
from fsspec.implementations.local import LocalFileSystem
from fsspec.registry import _registry as fsspec_registry_private
from fsspec.registry import known_implementations as fsspec_known_implementations
from fsspec.registry import register_implementation as fsspec_register_implementation
from fsspec.registry import registry as fsspec_registry

from upath import UPath
from upath.registry import available_implementations
from upath.registry import get_upath_class
from upath.registry import register_implementation

IMPLEMENTATIONS = {
    "abfs",
    "abfss",
    "adl",
    "az",
    "data",
    "file",
    "ftp",
    "gcs",
    "gs",
    "hdfs",
    "hf",
    "http",
    "https",
    "local",
    "memory",
    "s3",
    "s3a",
    "simplecache",
    "sftp",
    "smb",
    "ssh",
    "webdav",
    "webdav+http",
    "webdav+https",
    "github",
    "zip",
    "tar",
}


@pytest.fixture(autouse=True)
def reset_registry():
    from upath.registry import _registry

    try:
        yield
    finally:
        _registry._m.maps[0].clear()  # type: ignore


@pytest.fixture()
def fake_entrypoint():
    from importlib.metadata import EntryPoint

    from upath.registry import _registry

    ep = EntryPoint(
        name="myeps",
        value="upath.core:UPath",
        group="universal_pathlib.implementations",
    )
    old_registry = _registry._entries.copy()

    try:
        _registry._entries["myeps"] = ep
        yield
    finally:
        _registry._entries.clear()
        _registry._entries.update(old_registry)


def test_available_implementations():
    impl = available_implementations()
    assert len(impl) == len(set(impl))
    assert set(impl) == IMPLEMENTATIONS


@pytest.fixture
def fake_registered_proto():
    fake_proto = "".join(random.choices(string.ascii_lowercase, k=8))

    class FakeRandomFS(LocalFileSystem):
        protocol = fake_proto

    fsspec_register_implementation(fake_proto, FakeRandomFS)
    try:
        yield fake_proto
    finally:
        fsspec_registry_private.pop(fake_proto, None)


def test_available_implementations_with_fallback(fake_registered_proto):
    impl = available_implementations(fallback=True)
    assert fake_registered_proto in impl
    assert set(impl) == IMPLEMENTATIONS.union(
        {
            *fsspec_known_implementations,
            *fsspec_registry,
        }
    )


def test_available_implementations_with_entrypoint(fake_entrypoint):
    impl = available_implementations()
    assert set(impl) == IMPLEMENTATIONS.union({"myeps"})


def test_register_implementation():
    class MyProtoPath(UPath):
        pass

    register_implementation("myproto", MyProtoPath)

    assert get_upath_class("myproto") is MyProtoPath


def test_register_implementation_wrong_input():
    with pytest.raises(TypeError):
        register_implementation(None, UPath)  # type: ignore
    with pytest.raises(ValueError):
        register_implementation("incorrect**protocol", UPath)
    with pytest.raises(ValueError):
        register_implementation("myproto", object, clobber=True)  # type: ignore
    with pytest.raises(ValueError):
        register_implementation("file", UPath, clobber=False)
    assert set(available_implementations()) == IMPLEMENTATIONS


@pytest.mark.parametrize("protocol", IMPLEMENTATIONS)
def test_get_upath_class(protocol):
    upath_cls = get_upath_class("file")
    assert issubclass(upath_cls, UPath)


def test_get_upath_class_without_implementation(clear_registry):
    with pytest.warns(
        UserWarning, match="UPath 'mock' filesystem not explicitly implemented."
    ):
        upath_cls = get_upath_class("mock")
    assert issubclass(upath_cls, UPath)


def test_get_upath_class_without_implementation_no_fallback(clear_registry):
    assert get_upath_class("mock", fallback=False) is None


def test_get_upath_class_unknown_protocol(clear_registry):
    assert get_upath_class("doesnotexist") is None


def test_get_upath_class_from_entrypoint(fake_entrypoint):
    assert issubclass(get_upath_class("myeps"), UPath)


@pytest.mark.parametrize(
    "protocol", [pytest.param("", id="empty-str"), pytest.param(None, id="none")]
)
def test_get_upath_class_falsey_protocol(protocol):
    assert issubclass(get_upath_class(protocol), UPath)



from pathlib import Path

import pytest

from upath import UPath
from upath.implementations.webdav import WebdavPath

from ..cases import BaseTests
from ..utils import OverrideMeta
from ..utils import extends_base
from ..utils import overrides_base


class TestUPathWebdav(BaseTests, metaclass=OverrideMeta):
    @pytest.fixture(autouse=True, scope="function")
    def path(self, webdav_fixture):
        self.path = UPath(webdav_fixture, auth=("USER", "PASSWORD"))

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, WebdavPath)

    @extends_base
    def test_storage_options_base_url(self):
        # ensure that base_url is correct
        base_url = self.path.storage_options["base_url"]
        assert base_url == self.path.fs.client.base_url

    @overrides_base
    @pytest.mark.parametrize(
        "target_factory",
        [
            lambda obj, name: str(obj.joinpath(name).absolute()),
            pytest.param(
                lambda obj, name: UPath(obj.absolute().joinpath(name).path),
                marks=pytest.mark.xfail(reason="webdav has no root..."),
            ),
            pytest.param(
                lambda obj, name: Path(obj.absolute().joinpath(name).path),
                marks=pytest.mark.xfail(reason="webdav has no root..."),
            ),
            lambda obj, name: obj.absolute().joinpath(name),
        ],
        ids=[
            "str_absolute",
            "plain_upath_absolute",
            "plain_path_absolute",
            "self_upath_absolute",
        ],
    )
    def test_rename_with_target_absolute(self, target_factory):
        super().test_rename_with_target_absolute(target_factory)



import stat

import pytest

from upath import UnsupportedOperation
from upath import UPath
from upath.implementations.data import DataPath

from ..cases import JoinablePathTests
from ..cases import NonWritablePathTests
from ..cases import ReadablePathTests
from ..utils import OverrideMeta
from ..utils import overrides_base


class TestUPathDataPath(
    JoinablePathTests,
    ReadablePathTests,
    NonWritablePathTests,
    metaclass=OverrideMeta,
):
    """
    Unit-tests for the DataPath implementation of UPath.
    """

    @pytest.fixture(autouse=True)
    def path(self):
        path = "data:text/plain;base64,aGVsbG8gd29ybGQ="
        self.path = UPath(path)

    @pytest.fixture(autouse=True)
    def path_file(self, path):
        self.path_file = self.path

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, DataPath)

    @overrides_base
    def test_with_segments(self):
        # DataPath does not support joins, so in all usual cases it'll raise
        with pytest.raises(UnsupportedOperation):
            self.path.with_segments("data:text/plain;base64,", "aGVsbG8K")
        # but you can instantiate with a single full url
        self.path.with_segments("data:text/plain;base64,aGVsbG8K")

    @overrides_base
    def test_parents(self):
        # DataPath is always a absolute path with no parents
        assert self.path.parents == []

    @overrides_base
    def test_with_name(self):
        # DataPath does not support name changes
        with pytest.raises(UnsupportedOperation):
            self.path.with_name("newname")

    @overrides_base
    def test_with_suffix(self):
        # DataPath does not support suffix changes
        with pytest.raises(UnsupportedOperation):
            self.path.with_suffix(".new")

    @overrides_base
    def test_with_stem(self):
        # DataPath does not support stem changes
        with pytest.raises(UnsupportedOperation):
            self.path.with_stem("newname")

    @overrides_base
    def test_suffix(self):
        # DataPath does not have suffixes
        assert self.path.suffix == ""

    @overrides_base
    def test_suffixes(self):
        # DataPath does not have suffixes
        assert self.path.suffixes == []

    @overrides_base
    def test_repr_after_with_name(self):
        with pytest.raises(UnsupportedOperation):
            repr(self.path.with_name("data:,ABC"))

    @overrides_base
    def test_repr_after_with_suffix(self):
        with pytest.raises(UnsupportedOperation):
            repr(self.path.with_suffix(""))

    @overrides_base
    def test_child_path(self):
        # DataPath does not support joins, so child paths are unsupported
        with pytest.raises(UnsupportedOperation):
            super().test_child_path()

    @overrides_base
    def test_pickling_child_path(self):
        # DataPath does not support joins, so child paths are unsupported
        with pytest.raises(UnsupportedOperation):
            super().test_pickling_child_path()

    @overrides_base
    def test_relative_to(self):
        # DataPath only relative_to with itself
        with pytest.raises(ValueError):
            self.path.relative_to("data:,ABC")
        self.path.relative_to(self.path)

    @overrides_base
    def test_is_relative_to(self):
        # DataPath only relative_to with itself
        assert not self.path.is_relative_to("data:,ABC")
        assert self.path.is_relative_to(self.path)

    @overrides_base
    def test_full_match(self):
        assert self.path.full_match("*")
        assert not self.path.full_match("xxx")

    @overrides_base
    def test_trailing_slash_joinpath_is_identical(self):
        # DataPath has no slashes, and is not joinable
        with pytest.raises(UnsupportedOperation):
            super().test_trailing_slash_joinpath_is_identical()

    @overrides_base
    def test_trailing_slash_is_stripped(self):
        # DataPath has no slashes, and is not joinable
        with pytest.raises(UnsupportedOperation):
            super().test_trailing_slash_is_stripped()

    @overrides_base
    def test_parents_end_at_anchor(self):
        # DataPath does not support joins
        with pytest.raises(UnsupportedOperation):
            super().test_parents_end_at_anchor()

    @overrides_base
    def test_anchor_is_its_own_parent(self):
        # DataPath does not support joins
        assert self.path.path == self.path.parent.path

    @overrides_base
    def test_private_url_attr_in_sync(self):
        # DataPath does not support joins, so we check on self.path
        assert self.path._url

    @overrides_base
    def test_stat_dir_st_mode(self):
        # DataPath does not have directories
        assert not stat.S_ISDIR(self.path.stat().st_mode)

    @overrides_base
    def test_exists(self):
        # A valid DataPath always exists
        assert self.path.exists()

    @overrides_base
    def test_glob(self):
        # DataPath does not have dirs, joins or globs
        assert list(self.path.glob("*")) == []

    @overrides_base
    def test_rglob(self):
        # DataPath does not have dirs, joins or globs
        assert list(self.path.rglob("*")) == []

    @overrides_base
    def test_is_dir(self):
        # DataPath does not have directories
        assert not self.path.is_dir()

    @overrides_base
    def test_is_file(self):
        # DataPath is always a file
        assert self.path.is_file()

    @overrides_base
    def test_iterdir(self):
        # DataPath does not have directories
        with pytest.raises(NotADirectoryError):
            self.path.iterdir()

    @overrides_base
    def test_iterdir_parent_iteration(self):
        with pytest.raises(NotADirectoryError):
            super().test_iterdir_parent_iteration()

    @overrides_base
    def test_iterdir2(self):
        # DataPath does not have directories, or joins
        with pytest.raises(NotADirectoryError):
            self.path_file.iterdir()

    @overrides_base
    def test_iterdir_trailing_slash(self):
        # DataPath does not have directories, or joins
        with pytest.raises(UnsupportedOperation):
            super().test_iterdir_trailing_slash()

    @overrides_base
    def test_read_bytes(self):
        assert self.path.read_bytes() == b"hello world"

    @overrides_base
    def test_read_text(self):
        assert self.path.read_text() == "hello world"

    @overrides_base
    def test_walk(self):
        # DataPath does not have directories
        assert list(self.path.walk()) == []

    @overrides_base
    def test_walk_top_down_false(self):
        # DataPath does not have directories
        assert list(self.path.walk(top_down=False)) == []

    @overrides_base
    def test_samefile(self):
        # DataPath doesn't have joins, so only identical paths are samefile
        f1 = UPath("data:text/plain;base64,aGVsbG8gd29ybGQ=")
        f2 = UPath("data:text/plain;base64,SGVsbG8gd29ybGQ=")

        assert f1.samefile(f2) is False
        assert f1.samefile(f2.path) is False
        assert f1.samefile(f1) is True
        assert f1.samefile(f1.path) is True

    @overrides_base
    def test_info(self):
        # DataPath info checks
        p0 = self.path

        assert p0.info.exists() is True
        assert p0.info.is_file() is True
        assert p0.info.is_dir() is False
        assert p0.info.is_symlink() is False

    @overrides_base
    def test_mkdir_raises(self):
        # DataPaths always exist and are files
        with pytest.raises(FileExistsError):
            self.path_file.mkdir()

    @overrides_base
    def test_touch_raises(self):
        # DataPaths always exist, so touch is a noop
        self.path_file.touch()

    @overrides_base
    def test_unlink(self):
        # DataPaths can't be deleted
        with pytest.raises(UnsupportedOperation):
            self.path_file.unlink()

    @overrides_base
    def test_copy_into__dir_to_str_tempdir(self):
        # There are no directories in DataPath
        assert not self.path.is_dir()



import pytest

from upath import UPath
from upath.implementations.ftp import FTPPath
from upath.tests.cases import BaseTests
from upath.tests.utils import skip_on_windows

from ..utils import OverrideMeta
from ..utils import extends_base
from ..utils import overrides_base


@skip_on_windows
class TestUPathFTP(BaseTests, metaclass=OverrideMeta):

    @pytest.fixture(autouse=True)
    def path(self, ftp_server):
        self.path = UPath("", protocol="ftp", **ftp_server)
        self.prepare_file_system()

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, FTPPath)

    @extends_base
    def test_ftp_path_mtime(self, ftp_server):
        path = UPath("file1.txt", protocol="ftp", **ftp_server)
        path.touch()
        mtime = path.stat().st_mtime
        assert isinstance(mtime, float)



import pytest  # noqa: F401
from fsspec import get_filesystem_class

from upath import UPath
from upath.implementations.http import HTTPPath

from ..cases import JoinablePathTests
from ..cases import NonWritablePathTests
from ..cases import ReadablePathTests
from ..utils import OverrideMeta
from ..utils import extends_base
from ..utils import overrides_base
from ..utils import skip_on_windows
from ..utils import xfail_if_no_ssl_connection

try:
    get_filesystem_class("http")
except ImportError:
    pytestmark = pytest.mark.skip


@pytest.fixture
def internet_connection():
    import requests

    try:
        requests.get("http://example.com")
    except requests.exceptions.ConnectionError:
        pytest.xfail(reason="No internet connection")
    else:
        yield


def test_httppath(internet_connection):
    path = UPath("http://example.com")
    assert isinstance(path, HTTPPath)
    assert path.exists()


@xfail_if_no_ssl_connection
def test_httpspath(internet_connection):
    path = UPath("https://example.com")
    assert isinstance(path, HTTPPath)
    assert path.exists()


@skip_on_windows
class TestUPathHttp(
    JoinablePathTests,
    ReadablePathTests,
    NonWritablePathTests,
    metaclass=OverrideMeta,
):
    @pytest.fixture(autouse=True, scope="function")
    def path(self, http_fixture):
        self.path = UPath(http_fixture)

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, HTTPPath)

    @extends_base
    def test_work_at_root(self):
        assert "folder" in (f.name for f in self.path.parent.iterdir())

    @extends_base
    def test_resolve(self):
        # Also tests following redirects, because the test server issues a
        # 301 redirect for `http://127.0.0.1:8080/folder` to
        # `http://127.0.0.1:8080/folder/`
        assert str(self.path.resolve()).endswith("/")

    @overrides_base
    def test_info(self):
        # HTTPPath folders are files too

        p0 = self.path.joinpath("file1.txt")
        p1 = self.path.joinpath("folder1")

        assert p0.info.exists() is True
        assert p0.info.is_file() is True
        assert p0.info.is_dir() is False
        assert p0.info.is_symlink() is False
        assert p1.info.exists() is True
        assert (
            p1.info.is_file() is True
        )  # Weird quirk of how directories work in http fsspec
        assert p1.info.is_dir() is True
        assert p1.info.is_symlink() is False


@pytest.mark.parametrize(
    "args,parts",
    [
        (("http://example.com/"), ("http://example.com/", "")),
        (("http://example.com//"), ("http://example.com/", "", "")),
        (("http://example.com///"), ("http://example.com/", "", "", "")),
        (("http://example.com/a"), ("http://example.com/", "a")),
        (("http://example.com/a/"), ("http://example.com/", "a", "")),
        (("http://example.com/a/b"), ("http://example.com/", "a", "b")),
        (("http://example.com/a//b"), ("http://example.com/", "a", "", "b")),
        (("http://example.com/a//b/"), ("http://example.com/", "a", "", "b", "")),
    ],
)
def test_empty_parts(args, parts):
    pth = UPath(args)
    pth_parts = pth.parts
    assert pth_parts == parts


def test_query_parameters_passthrough():
    pth = UPath("http://example.com/?a=1&b=2")
    assert pth.parts == ("http://example.com/", "?a=1&b=2")


@pytest.mark.parametrize(
    "base,rel,expected",
    [
        (
            "http://www.example.com/a/b/index.html",
            "image.png?version=1",
            "http://www.example.com/a/b/image.png?version=1",
        ),
        (
            "http://www.example.com/a/b/index.html",
            "../image.png",
            "http://www.example.com/a/image.png",
        ),
        (
            "http://www.example.com/a/b/index.html",
            "/image.png",
            "http://www.example.com/image.png",
        ),
        (
            "http://www.example.com/a/b/index.html",
            "sftp://other.com/image.png",
            "sftp://other.com/image.png",
        ),
        (
            "http://www.example.com/a/b/index.html",
            "//other.com/image.png",
            "http://other.com/image.png",
        ),
    ],
)
def test_joinuri_behavior(base, rel, expected):
    p0 = UPath(base)
    pr = p0.joinuri(rel)
    pe = UPath(expected)
    assert pr == pe


NORMALIZATIONS = (
    ("unnormalized", "normalized"),
    (
        # Expected normalization results according to curl
        ("http://example.com", "http://example.com/"),
        ("http://example.com/", "http://example.com/"),
        ("http://example.com/a", "http://example.com/a"),
        ("http://example.com//a", "http://example.com//a"),
        ("http://example.com///a", "http://example.com///a"),
        ("http://example.com////a", "http://example.com////a"),
        ("http://example.com/a/.", "http://example.com/a/"),
        ("http://example.com/a/./", "http://example.com/a/"),
        ("http://example.com/a/./b", "http://example.com/a/b"),
        ("http://example.com/a/.//", "http://example.com/a//"),
        ("http://example.com/a/.//b", "http://example.com/a//b"),
        ("http://example.com/a//.", "http://example.com/a//"),
        ("http://example.com/a//./", "http://example.com/a//"),
        ("http://example.com/a//./b", "http://example.com/a//b"),
        ("http://example.com/a//.//", "http://example.com/a///"),
        ("http://example.com/a//.//b", "http://example.com/a///b"),
        ("http://example.com/a/..", "http://example.com/"),
        ("http://example.com/a/../", "http://example.com/"),
        ("http://example.com/a/../.", "http://example.com/"),
        ("http://example.com/a/../..", "http://example.com/"),
        ("http://example.com/a/../../", "http://example.com/"),
        ("http://example.com/a/../..//", "http://example.com//"),
        ("http://example.com/a/..//", "http://example.com//"),
        ("http://example.com/a/..//.", "http://example.com//"),
        ("http://example.com/a/..//..", "http://example.com/"),
        ("http://example.com/a/../b", "http://example.com/b"),
        ("http://example.com/a/..//b", "http://example.com//b"),
        ("http://example.com/a//..", "http://example.com/a/"),
        ("http://example.com/a//../", "http://example.com/a/"),
        ("http://example.com/a//../.", "http://example.com/a/"),
        ("http://example.com/a//../..", "http://example.com/"),
        ("http://example.com/a//../../", "http://example.com/"),
        ("http://example.com/a//../..//", "http://example.com//"),
        ("http://example.com/a//..//..", "http://example.com/a/"),
        ("http://example.com/a//../b", "http://example.com/a/b"),
        ("http://example.com/a//..//", "http://example.com/a//"),
        ("http://example.com/a//..//.", "http://example.com/a//"),
        ("http://example.com/a//..//b", "http://example.com/a//b"),
    ),
)


@pytest.mark.parametrize(*NORMALIZATIONS)
def test_normalize(unnormalized, normalized):
    expected = HTTPPath(normalized)
    pth = HTTPPath(unnormalized)
    assert expected.protocol in {"http", "https"}
    assert pth.protocol in {"http", "https"}

    # Normalise only, do not attempt to follow redirects for http:// paths here
    result = pth.resolve(strict=True, follow_redirects=False)

    str_expected = str(expected)
    str_result = str(result)
    assert expected == result
    assert str_expected == str_result



"""
GitHub file system implementation
"""

from __future__ import annotations

import sys
from collections.abc import Sequence
from typing import TYPE_CHECKING

from upath.core import UnsupportedOperation
from upath.core import UPath
from upath.types import JoinablePathLike

if TYPE_CHECKING:
    from typing import Literal

    if sys.version_info >= (3, 11):
        from typing import Unpack
    else:
        from typing_extensions import Unpack

    from upath._chain import FSSpecChainParser
    from upath.types.storage_options import GitHubStorageOptions

__all__ = ["GitHubPath"]


class GitHubPath(UPath):
    """
    GitHubPath supporting the fsspec.GitHubFileSystem
    """

    __slots__ = ()

    if TYPE_CHECKING:

        def __init__(
            self,
            *args: JoinablePathLike,
            protocol: Literal["github"] | None = ...,
            chain_parser: FSSpecChainParser = ...,
            **storage_options: Unpack[GitHubStorageOptions],
        ) -> None: ...

    @property
    def path(self) -> str:
        pth = super().path
        if pth == ".":
            return ""
        return pth

    @property
    def parts(self) -> Sequence[str]:
        parts = super().parts
        if parts and parts[0] == "/":
            return parts[1:]
        else:
            return parts

    def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
        raise UnsupportedOperation

    def mkdir(
        self,
        mode: int = 0o777,
        parents: bool = False,
        exist_ok: bool = False,
    ) -> None:
        raise UnsupportedOperation

    def unlink(self, missing_ok: bool = False) -> None:
        raise UnsupportedOperation

    def write_bytes(self, data: bytes) -> int:
        raise UnsupportedOperation

    def write_text(
        self,
        data: str,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> int:
        raise UnsupportedOperation("GitHubPath does not support writing")



"""Generates the _flavour_sources.py file"""

from __future__ import annotations

import inspect
import re
import sys
import warnings
from io import StringIO
from typing import Any
from unittest.mock import Mock

from fsspec.registry import available_protocols
from fsspec.registry import get_filesystem_class
from fsspec.spec import AbstractFileSystem
from fsspec.utils import get_package_version_without_import

HEADER = '''\
""" upath._flavour_sources

<experimental!>

Warning
-------
  Do not modify this file manually!
  It is generated by `dev/generate_flavours.py`

To be able to parse the different filesystem uri schemes, we need
the string parsing functionality each of the filesystem implementations.
In an attempt to support parsing uris without having to import the
specific filesystems, we extract the necessary subset of the
AbstractFileSystem classes and generate a new "flavour" class for
each of the known filesystems. This will allow us to provide a
`PurePath` equivalent `PureUPath` for each protocol in the future
without a direct dependency on the underlying filesystem package.

"""
'''

IMPORTS = """\
from __future__ import annotations

import logging
import os
import re
from pathlib import PurePath
from pathlib import PureWindowsPath
from typing import Any
from typing import Literal
from typing import cast
from urllib.parse import parse_qs
from urllib.parse import urlsplit

from fsspec.implementations.local import make_path_posix
from fsspec.utils import infer_storage_options
from fsspec.utils import stringify_path

"""

INIT_CODE = '''\
__all__ = [
    "AbstractFileSystemFlavour",
    "FileSystemFlavourBase",
    "flavour_registry",
]

logger = logging.getLogger(__name__)
flavour_registry: dict[str, type[FileSystemFlavourBase]] = {}


class FileSystemFlavourBase:
    """base class for the fsspec flavours"""

    protocol: str | tuple[str, ...]
    root_marker: Literal["/", ""]
    sep: Literal["/"]

    @classmethod
    def _strip_protocol(cls, path):
        raise NotImplementedError

    @staticmethod
    def _get_kwargs_from_urls(path):
        raise NotImplementedError

    @classmethod
    def _parent(cls, path):
        raise NotImplementedError

    def __init_subclass__(cls: Any, **kwargs):
        if isinstance(cls.protocol, str):
            protocols = (cls.protocol,)
        else:
            protocols = tuple(cls.protocol)
        for protocol in protocols:
            if protocol in flavour_registry:
                raise ValueError(f"protocol {protocol!r} already registered")
            flavour_registry[protocol] = cls
'''

BASE_CLASS_NAME_SUFFIX = "Flavour"
BASE_CLASS_NAME = f"{AbstractFileSystem.__name__}{BASE_CLASS_NAME_SUFFIX}"

SKIP_PROTOCOLS = [
    "dir",
    "blockcache",
    "cached",
    "filecache",
]

FIX_PROTOCOLS = {
    "MemFS": ("memfs",),
    "AsyncLocalFileSystem": (),
}

FIX_METHODS = {
    "GCSFileSystem": ["_strip_protocol", "_get_kwargs_from_urls", "_split_path"],
    "SimpleCacheFileSystem": [],
}


def _fix_abstract_file_system(x: str) -> str:
    x = re.sub(
        "protocol = 'abstract'", "protocol: str | tuple[str, ...] = 'abstract'", x
    )
    x = re.sub("root_marker = ''", "root_marker: Literal['', '/'] = ''", x)
    x = re.sub("sep = '/'", "sep: Literal['/'] = '/'", x)
    return x


def _fix_azure_blob_file_system(x: str) -> str:
    x = re.sub(
        r"if isinstance\(path, list\):",
        "if isinstance(path, list):  # type: ignore[unreachable]",
        x,
    )
    x = re.sub(
        r"(return \[.*\])",
        r"\1  # type: ignore[unreachable]",
        x,
    )
    return x


def _fix_data_file_system(x: str) -> str:
    return re.sub(
        "sep = '/'",
        "sep = ''  # type: ignore[assignment]\n    "
        "altsep = ' '  # type: ignore[assignment]",
        x,
    )


def _fix_memfs_file_system(x: str) -> str:
    return re.sub(
        "_MemFS",
        "MemoryFileSystemFlavour",
        x,
    )


def _fix_memory_file_system(x: str) -> str:
    return re.sub(
        "LocalFileSystem",
        "LocalFileSystemFlavour",
        x,
    )


def _fix_oss_file_system(x: str) -> str:
    x = re.sub(
        r"path_string: str = stringify_path\(path\)",
        "path_string = stringify_path(path)",
        x,
    )
    return x


def _fix_xrootd_file_system(x: str) -> str:
    x = re.sub(
        r"return client[.]URL\(path\)[.]path_with_params",
        "x = urlsplit(path); return (x.path + f'?{x.query}' if x.query else '')",
        x,
    )
    x = re.sub(r"client[.]URL\(u\)", "urlsplit(u)", x)
    return re.sub(
        "url.hostid",
        "url.netloc",
        x,
    )


FIX_SOURCE = {
    "AbstractFileSystem": _fix_abstract_file_system,
    "AzureBlobFileSystem": _fix_azure_blob_file_system,
    "DataFileSystem": _fix_data_file_system,
    "MemFS": _fix_memfs_file_system,
    "MemoryFileSystem": _fix_memory_file_system,
    "OSSFileSystem": _fix_oss_file_system,
    "XRootDFileSystem": _fix_xrootd_file_system,
}


def before_imports() -> None:
    """allow to patch the generated state before importing anything"""
    # patch libarchive
    sys.modules["libarchive"] = Mock()
    sys.modules["libarchive.ffi"] = Mock()
    # patch xrootd
    sys.modules["XRootD"] = Mock()
    sys.modules["XRootD.client"] = Mock()
    sys.modules["XRootD.client.flags"] = Mock()
    sys.modules["XRootD.client.responses"] = Mock()


def get_protos(cls: type, remove: str, add: str) -> tuple[str, ...]:
    try:
        return FIX_PROTOCOLS[cls.__name__]
    except KeyError:
        pass
    if isinstance(cls.protocol, str):
        p = [cls.protocol, add]
    else:
        p = [*cls.protocol, add]
    return tuple([x for x in p if x != remove])


def get_fsspec_filesystems_and_protocol_errors() -> (
    tuple[dict[type[AbstractFileSystem], tuple[str, ...]], dict[str, str]]
):
    before_imports()

    classes: dict[type[AbstractFileSystem], tuple[str]] = {}
    errors: dict[str, str] = {}

    for protocol in available_protocols():
        if protocol in SKIP_PROTOCOLS:
            continue
        try:
            cls = get_filesystem_class(protocol)
        except ImportError as err:
            errors[protocol] = str(err)
        else:
            protos = get_protos(cls, remove="abstract", add=protocol)
            cprotos = classes.get(cls, [])
            classes[cls] = tuple(dict.fromkeys([*cprotos, *protos]))
    return classes, errors


def _get_plain_method(cls, name):
    for c in cls.__mro__:
        try:
            return c.__dict__[name]
        except KeyError:
            pass
    else:
        raise AttributeError(f"{cls.__name__}.{name} not found")


def get_subclass_methods(cls: type) -> list[str]:  # noqa: C901
    try:
        return FIX_METHODS[cls.__name__]
    except KeyError:
        pass
    errors = []

    # storage options
    so = None
    base_get_kwargs_from_urls = _get_plain_method(
        AbstractFileSystem, "_get_kwargs_from_urls"
    )
    try:
        cls_get_kwargs_from_urls = _get_plain_method(cls, "_get_kwargs_from_urls")
    except AttributeError:
        errors.append("missing `_get_kwargs_from_urls()`")
    else:
        so = cls_get_kwargs_from_urls is base_get_kwargs_from_urls
        if not isinstance(cls_get_kwargs_from_urls, staticmethod):
            warnings.warn(
                f"{cls.__name__}: {cls_get_kwargs_from_urls!r} not a staticmethod",
                RuntimeWarning,
                stacklevel=2,
            )

    # strip protocol
    sp = None
    base_strip_protocol = _get_plain_method(AbstractFileSystem, "_strip_protocol")
    try:
        cls_strip_protocol = _get_plain_method(cls, "_strip_protocol")
    except AttributeError:
        errors.append("missing `_strip_protocol()`")
    else:
        if isinstance(cls_strip_protocol, staticmethod):
            warnings.warn(
                f"{cls.__name__}: {cls_strip_protocol.__name__!r} is not a classmethod",
                UserWarning,
                stacklevel=2,
            )
            sp = False
        elif isinstance(cls_strip_protocol, classmethod):
            sp = cls_strip_protocol.__func__ is base_strip_protocol.__func__
        else:
            errors.append(
                f"{cls.__name__}: {cls_strip_protocol.__name__!r} not a classmethod"
            )

    # _parent
    pt = None
    base_parent = _get_plain_method(AbstractFileSystem, "_parent")
    try:
        cls_parent = _get_plain_method(cls, "_parent")
    except AttributeError:
        errors.append("missing `_parent()`")
    else:
        pt = cls_parent is base_parent

    if errors or sp is None or so is None:
        raise AttributeError(" AND ".join(errors))

    methods = []
    if not sp:
        methods.append("_strip_protocol")
    if not so:
        methods.append("_get_kwargs_from_urls")
    if not pt:
        methods.append("_parent")
    return methods


def generate_class_source_code(
    cls: type,
    methods: list[str],
    overrides: dict[str, Any],
    attributes: list[str],
    cls_suffix: str,
    base_cls: str | None,
) -> str:
    s = ["\n"]
    if base_cls:
        s += [f"class {cls.__name__}{cls_suffix}({base_cls}):"]
    else:
        s += [f"class {cls.__name__}{cls_suffix}:"]
    mod_ver = get_package_version_without_import(cls.__module__.partition(".")[0])
    s.append(f"    __orig_class__ = '{cls.__module__}.{cls.__name__}'")
    s.append(f"    __orig_version__ = {mod_ver!r}")
    for attr, value in overrides.items():
        s.append(f"    {attr} = {value!r}")
    for attr in attributes:
        s.append(f"    {attr} = {getattr(cls, attr)!r}")
    if getattr(cls, "local_file", False):
        s.append("    local_file = True")
    s.append("")
    for method in methods:
        s.append(inspect.getsource(getattr(cls, method)))
    try:
        fix_func = FIX_SOURCE[cls.__name__]
    except KeyError:
        return "\n".join(s)
    else:
        return "\n".join(fix_func(line) for line in s)


def create_source() -> str:
    buf = StringIO()
    buf.write(HEADER)

    classes, errors = get_fsspec_filesystems_and_protocol_errors()

    srcs = [
        generate_class_source_code(
            AbstractFileSystem,
            ["_strip_protocol", "_get_kwargs_from_urls", "_parent"],
            {},
            ["protocol", "root_marker", "sep"],
            cls_suffix=BASE_CLASS_NAME_SUFFIX,
            base_cls="FileSystemFlavourBase",
        )
    ]

    for cls in sorted(classes, key=lambda cls: cls.__name__):
        try:
            sub_cls_methods = get_subclass_methods(cls)
        except AttributeError as err:
            protos = (cls.protocol,) if isinstance(cls.protocol, str) else cls.protocol
            for proto in protos:
                errors[proto] = str(err)
            continue
        sub_cls = generate_class_source_code(
            cls,
            sub_cls_methods,
            {"protocol": classes[cls]},
            ["root_marker", "sep"],
            cls_suffix=BASE_CLASS_NAME_SUFFIX,
            base_cls=BASE_CLASS_NAME,
        )
        srcs.append(sub_cls)

    if SKIP_PROTOCOLS:
        buf.write("#\n# skipping protocols:\n")
        for protocol in sorted(SKIP_PROTOCOLS):
            buf.write(f"#   - {protocol}\n")

    if errors:
        buf.write("# protocol import errors:\n")
        for protocol, error_msg in sorted(errors.items()):
            buf.write(f"#   - {protocol} ({error_msg})\n")
        buf.write("#\n")

    buf.write(IMPORTS)
    buf.write(INIT_CODE)
    for cls_src in srcs:
        buf.write(cls_src)

    return buf.getvalue().removesuffix("\n")


if __name__ == "__main__":
    print(create_source())



import functools
import os
import platform
import sys

import pytest

from upath import UPath
from upath.implementations.github import GitHubPath

from ..cases import JoinablePathTests
from ..cases import NonWritablePathTests
from ..cases import ReadablePathTests
from ..utils import OverrideMeta
from ..utils import overrides_base

pytestmark = pytest.mark.skipif(
    os.environ.get("CI", False)
    and not (
        platform.system() == "Linux" and sys.version_info[:2] in {(3, 9), (3, 13)}
    ),
    reason="Skipping GitHubPath tests to prevent rate limiting on GitHub API.",
)


def xfail_on_github_connection_error(func):
    """Method decorator to xfail tests on GitHub rate limit or connection errors."""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except Exception as e:
            str_e = str(e)
            if "rate limit exceeded" in str_e or "too many requests for url" in str_e:
                pytest.xfail("GitHub API rate limit exceeded")
            elif (
                "nodename nor servname provided, or not known" in str_e
                or "Network is unreachable" in str_e
                or "NameResolutionError" in str_e
            ):
                pytest.xfail("No internet connection")
            else:
                raise

    return wrapper


def wrap_all_tests(decorator):
    """Class decorator factory to wrap all test methods with a given decorator."""

    def class_decorator(cls):
        for attr_name in dir(cls):
            if attr_name.startswith("test_"):
                orig_method = getattr(cls, attr_name)
                setattr(cls, attr_name, decorator(orig_method))
        return cls

    return class_decorator


@wrap_all_tests(xfail_on_github_connection_error)
class TestUPathGitHubPath(
    JoinablePathTests,
    ReadablePathTests,
    NonWritablePathTests,
    metaclass=OverrideMeta,
):
    """
    Unit-tests for the GitHubPath implementation of UPath.
    """

    @pytest.fixture(autouse=True)
    def path(self):
        """
        Fixture for the UPath instance to be tested.
        """
        path = "github://ap--:universal_pathlib@test_data/data"
        self.path = UPath(path)

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, GitHubPath)



from __future__ import annotations

import sys
import warnings
from collections import defaultdict
from collections import deque
from collections.abc import Iterator
from collections.abc import MutableMapping
from collections.abc import Sequence
from collections.abc import Set
from itertools import zip_longest
from typing import TYPE_CHECKING
from typing import Any
from typing import NamedTuple

from upath._flavour import WrappedFileSystemFlavour
from upath._protocol import get_upath_protocol
from upath.registry import available_implementations
from upath.types import UNSET_DEFAULT

if TYPE_CHECKING:
    if sys.version_info >= (3, 11):
        from typing import Never
        from typing import Self
    else:
        from typing_extensions import Never
        from typing_extensions import Self

__all__ = [
    "ChainSegment",
    "Chain",
    "FSSpecChainParser",
    "DEFAULT_CHAIN_PARSER",
]


class ChainSegment(NamedTuple):
    path: str | None  # support for path passthrough (i.e. simplecache)
    protocol: str
    storage_options: dict[str, Any]


class Chain:
    """holds current chain segments"""

    __slots__ = (
        "_segments",
        "_index",
    )

    def __init__(
        self,
        *segments: ChainSegment,
        index: int = 0,
    ) -> None:
        if not (0 <= index < len(segments)):
            raise ValueError("index must be between 0 and len(segments)")
        self._segments = segments
        self._index = index

    def __repr__(self) -> str:
        args = ", ".join(map(repr, self._segments))
        if self._index != 0:
            args += f", index={self._index}"
        return f"{type(self).__name__}({args})"

    @property
    def current(self) -> ChainSegment:
        return self._segments[self._index]

    @property
    def _path_index(self) -> int:
        for idx, segment in enumerate(self._segments[self._index :], start=self._index):
            if segment.path is not None:
                return idx
        raise IndexError("No target path found")

    @property
    def active_path(self) -> str:
        path = self._segments[self._path_index].path
        if path is None:
            raise RuntimeError
        return path

    @property
    def active_path_protocol(self) -> str:
        return self._segments[self._path_index].protocol

    def replace(
        self,
        *,
        path: str | None = None,
        protocol: str | None = None,
        storage_options: dict[str, Any] | None = None,
    ) -> Self:
        """replace the current chain segment keeping remaining chain segments"""
        segments = self.to_list()
        index = self._index

        replacements: MutableMapping[int, dict[str, Any]] = defaultdict(dict)
        if protocol is not None:
            replacements[index]["protocol"] = protocol
        if storage_options is not None:
            replacements[index]["storage_options"] = storage_options
        if path is not None:
            replacements[self._path_index]["path"] = path

        for idx, items in replacements.items():
            segments[idx] = segments[idx]._replace(**items)

        return type(self)(*segments, index=index)

    def to_list(self) -> list[ChainSegment]:
        """return a list of chain segments unnesting target_* segments"""
        queue = deque(self._segments)
        segments = []
        while queue:
            segment = queue.popleft()
            if (
                "target_protocol" in segment.storage_options
                and "fo" in segment.storage_options
            ):
                storage_options = segment.storage_options.copy()
                target_options = storage_options.pop("target_options", {})
                target_protocol = storage_options.pop("target_protocol")
                fo = storage_options.pop("fo")
                queue.appendleft(ChainSegment(fo, target_protocol, target_options))
                segments.append(
                    ChainSegment(segment.path, segment.protocol, storage_options)
                )
            elif not segments or segment != segments[-1]:
                segments.append(segment)
        return segments

    @classmethod
    def from_list(cls, segments: list[ChainSegment], index: int = 0) -> Self:
        return cls(*segments, index=index)

    def nest(self) -> ChainSegment:
        """return a nested target_* structure"""
        # see: fsspec.core.url_to_fs
        inkwargs: dict[str, Any] = {}
        # Reverse iterate the chain, creating a nested target_* structure
        chain = self._segments
        _prev = chain[-1].path
        for i, ch in enumerate(reversed(chain)):
            urls, protocol, kw = ch
            if urls is None:
                urls = _prev
            _prev = urls
            if i == len(chain) - 1:
                inkwargs = {**kw, **inkwargs}
                continue
            inkwargs["target_options"] = {**kw, **inkwargs}
            inkwargs["target_protocol"] = protocol
            inkwargs["fo"] = urls  # codespell:ignore fo
        urlpath, protocol, _ = chain[0]
        return ChainSegment(urlpath, protocol, inkwargs)


def _iter_fileobject_protocol_options(
    fileobject: str | None,
    protocol: str,
    storage_options: dict[str, Any],
    /,
) -> Iterator[tuple[str | None, str, dict[str, Any]]]:
    """yields fileobject, protocol and remaining storage options"""
    so = storage_options.copy()
    while "target_protocol" in so:
        t_protocol = so.pop("target_protocol", "")
        t_fileobject = so.pop("fo", None)  # codespell:ignore fo
        t_so = so.pop("target_options", {})
        yield fileobject, protocol, so
        fileobject, protocol, so = t_fileobject, t_protocol, t_so
    yield fileobject, protocol, so


class FSSpecChainParser:
    """parse an fsspec chained urlpath"""

    def __init__(self) -> None:
        self.link: str = "::"
        self.known_protocols: Set[str] = set()

    def unchain(
        self,
        path: str,
        _deprecated_storage_options: Never = UNSET_DEFAULT,
        /,
        *,
        protocol: str | None = None,
        storage_options: dict[str, Any] | None = None,
    ) -> list[ChainSegment]:
        """implements same behavior as fsspec.core._un_chain

        two differences:
        1. it sets the urlpath to None for upstream filesystems that passthrough
        2. it checks against the known protocols for exact matches

        """
        if _deprecated_storage_options is not UNSET_DEFAULT:
            warnings.warn(
                "passing storage_options as positional argument is deprecated, "
                "pass as keyword argument instead",
                DeprecationWarning,
                stacklevel=2,
            )
            if storage_options is not None:
                raise ValueError(
                    "cannot pass storage_options both positionally and as keyword"
                )
            storage_options = _deprecated_storage_options
            protocol = protocol or storage_options.get("protocol")
        if storage_options is None:
            storage_options = {}

        segments: list[ChainSegment] = []
        path_bit: str | None
        next_path_overwrite: str | None = None
        for proto0, bit in zip_longest([protocol], path.split(self.link)):
            # get protocol and path_bit
            if (
                "://" in bit  # uri-like, fast-path (redundant)
                or "/" in bit  # path-like, fast-path
            ):
                proto = get_upath_protocol(bit, protocol=proto0)
                flavour = WrappedFileSystemFlavour.from_protocol(proto)
                path_bit = flavour.strip_protocol(bit)
                extra_so = flavour.get_kwargs_from_url(bit)
            elif bit in self.known_protocols and (
                proto0 is None or bit == proto0
            ):  # exact match a fsspec protocol
                proto = bit
                path_bit = None
                extra_so = {}
            elif bit in (m := set(available_implementations(fallback=True))) and (
                proto0 is None or bit == proto0
            ):
                self.known_protocols = m
                proto = bit
                path_bit = None
                extra_so = {}
            else:
                proto = get_upath_protocol(bit, protocol=proto0)
                flavour = WrappedFileSystemFlavour.from_protocol(proto)
                path_bit = flavour.strip_protocol(bit)
                extra_so = flavour.get_kwargs_from_url(bit)
            if proto in {"blockcache", "filecache", "simplecache"}:
                if path_bit is not None:
                    next_path_overwrite = path_bit or "/"
                path_bit = None
            elif next_path_overwrite is not None:
                path_bit = next_path_overwrite
                next_path_overwrite = None
            segments.append(ChainSegment(path_bit, proto, extra_so))

        root_so = segments[0].storage_options
        for segment, proto_fo_so in zip_longest(
            segments,
            _iter_fileobject_protocol_options(
                path_bit if segments else None,
                protocol or "",
                storage_options,
            ),
        ):
            t_fo, t_proto, t_so = proto_fo_so or (None, "", {})
            if segment is None:
                if next_path_overwrite is not None:
                    t_fo = next_path_overwrite
                    next_path_overwrite = None
                segments.append(ChainSegment(t_fo, t_proto, t_so))
            else:
                proto = segment.protocol
                # check if protocol is consistent with storage options
                if t_proto and t_proto != proto:
                    raise ValueError(
                        f"protocol {proto!r} collides with target_protocol {t_proto!r}"
                    )
                # update the storage_options
                segment.storage_options.update(root_so.pop(proto, {}))
                segment.storage_options.update(t_so)

        return segments

    def chain(self, segments: Sequence[ChainSegment]) -> tuple[str, dict[str, Any]]:
        """returns a chained urlpath from the segments"""
        urlpaths = []
        kwargs = {}
        for segment in segments:
            if segment.protocol and segment.path is not None:
                # FIXME: currently unstrip_protocol is only implemented by
                #   AbstractFileSystem, LocalFileSystem, and OSSFileSystem
                #   so to make this work we just implement it ourselves here.
                #   To do this properly we would need to instantiate the
                #   filesystem with its storage options and call
                #   fs.unstrip_protocol(segment.path)
                if segment.path.startswith(f"{segment.protocol}:/"):
                    urlpath = segment.path
                else:
                    urlpath = f"{segment.protocol}://{segment.path}"
            elif segment.protocol:
                urlpath = segment.protocol
            elif segment.path is not None:
                urlpath = segment.path
            else:
                warnings.warn(
                    f"skipping invalid segment {segment}",
                    RuntimeWarning,
                    stacklevel=2,
                )
                continue
            urlpaths.append(urlpath)
            # TODO: ensure roundtrip with unchain behavior
            if segment.storage_options:
                kwargs[segment.protocol] = segment.storage_options
        return self.link.join(urlpaths), kwargs


DEFAULT_CHAIN_PARSER = FSSpecChainParser()


if __name__ == "__main__":
    from pprint import pp

    from fsspec.core import _un_chain

    chained_path = "simplecache::zip://haha.csv::gcs://bucket/file.zip"
    chained_kw = {"zip": {"allowZip64": False}}
    print(chained_path, chained_kw)
    out0 = _un_chain(chained_path, chained_kw)
    out1 = FSSpecChainParser().unchain(chained_path, storage_options=chained_kw)

    pp(out0)
    pp(out1)

    rechained_path, rechained_kw = FSSpecChainParser().chain(out1)
    print(rechained_path, rechained_kw)

    # UPath should store segments and access the path to operate on
    # through segments.current.path
    segments0 = Chain.from_list(segments=out1, index=1)
    assert segments0.current.protocol == "zip"

    # try to switch out zip path
    segments1 = segments0.replace(path="/newfile.csv")
    new_path, new_kw = FSSpecChainParser().chain(segments1.to_list())
    print(new_path, new_kw)



from __future__ import annotations

import sys
from collections.abc import Mapping
from collections.abc import Sequence
from typing import TYPE_CHECKING
from typing import Any
from urllib.parse import urlsplit

from fsspec.registry import known_implementations
from fsspec.registry import register_implementation

from upath.core import UPath
from upath.types import JoinablePathLike

if TYPE_CHECKING:
    from typing import Literal

    if sys.version_info >= (3, 11):
        from typing import Unpack
    else:
        from typing_extensions import Unpack

    from upath._chain import FSSpecChainParser
    from upath.types.storage_options import WebdavStorageOptions

__all__ = ["WebdavPath"]

# webdav was only registered in fsspec>=2022.5.0
if "webdav" not in known_implementations:
    import webdav4.fsspec

    register_implementation("webdav", webdav4.fsspec.WebdavFileSystem)


class WebdavPath(UPath):
    __slots__ = ()

    if TYPE_CHECKING:

        def __init__(
            self,
            *args: JoinablePathLike,
            protocol: Literal["webdav+http", "webdav+https"] | None = ...,
            chain_parser: FSSpecChainParser = ...,
            **storage_options: Unpack[WebdavStorageOptions],
        ) -> None: ...

    @classmethod
    def _transform_init_args(
        cls,
        args: tuple[JoinablePathLike, ...],
        protocol: str,
        storage_options: dict[str, Any],
    ) -> tuple[tuple[JoinablePathLike, ...], str, dict[str, Any]]:
        if not args:
            args = ("/",)
        elif args and protocol in {"webdav+http", "webdav+https"}:
            args0, *argsN = args
            url = urlsplit(str(args0))
            base = url._replace(scheme=protocol.split("+")[1], path="").geturl()
            args0 = url._replace(scheme="", netloc="").geturl() or "/"
            storage_options["base_url"] = base
            args = (args0, *argsN)
        if "base_url" not in storage_options:
            raise ValueError(
                f"must provide `base_url` storage option for args: {args!r}"
            )
        return super()._transform_init_args(args, "webdav", storage_options)

    @classmethod
    def _parse_storage_options(
        cls,
        urlpath: str,
        protocol: str,
        storage_options: Mapping[str, Any],
    ) -> dict[str, Any]:
        so = dict(storage_options)
        if urlpath.startswith(("webdav+http:", "webdav+https:")):
            url = urlsplit(str(urlpath))
            base = url._replace(scheme=url.scheme.split("+")[1], path="").geturl()
            urlpath = url._replace(scheme="", netloc="").geturl() or "/"
            so.setdefault("base_url", base)
        return super()._parse_storage_options(urlpath, "webdav", so)

    @property
    def parts(self) -> Sequence[str]:
        parts = super().parts
        if parts and parts[0] == "/":
            return parts[1:]
        else:
            return parts



import os
import pathlib
import pickle
import sys
import warnings
from urllib.parse import SplitResult

import pathlib_abc
import pytest

from upath import UPath
from upath.implementations.cloud import GCSPath
from upath.implementations.cloud import S3Path
from upath.registry import get_upath_class
from upath.registry import register_implementation
from upath.types import ReadablePath
from upath.types import WritablePath

from .cases import BaseTests
from .utils import OverrideMeta
from .utils import only_on_windows
from .utils import overrides_base
from .utils import skip_on_windows


@skip_on_windows
def test_posix_path(local_testdir):
    assert isinstance(UPath(local_testdir), pathlib.PosixPath)


@only_on_windows
def test_windows_path(local_testdir):
    assert isinstance(UPath(local_testdir), pathlib.WindowsPath)


def test_UPath_untested_protocol_warning(clear_registry):
    with warnings.catch_warnings(record=True) as w:
        _ = UPath("mock:/")
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert "mock" in str(w[-1].message)


def test_UPath_file_protocol_no_warning():
    with warnings.catch_warnings(record=True) as w:
        _ = UPath("file:/")
        assert len(w) == 0


class TestUpath(BaseTests, metaclass=OverrideMeta):
    @pytest.fixture(autouse=True)
    def path(self, local_testdir):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # On Windows the path needs to be prefixed with `/`, because
            # `UPath` implements `_posix_flavour`, which requires a `/` root
            # in order to correctly deserialize pickled objects
            root = "/" if sys.platform.startswith("win") else ""
            self.path = UPath(f"mock:{root}{local_testdir}")

    @overrides_base
    def test_is_correct_class(self):
        # testing dynamically created UPath classes
        from upath.implementations._experimental import _MockPath

        assert isinstance(self.path, _MockPath)

    @overrides_base
    @pytest.mark.skipif(
        sys.platform.startswith("win"),
        reason="mock fs is not well defined on windows",
    )
    def test_parents_are_absolute(self):
        super().test_parents_are_absolute()

    @overrides_base
    @pytest.mark.skipif(
        sys.platform.startswith("win"),
        reason="mock fs is not well defined on windows",
    )
    def test_anchor_is_its_own_parent(self):
        super().test_anchor_is_its_own_parent()

    @overrides_base
    @pytest.mark.skipif(
        sys.platform.startswith("win"),
        reason="mock fs is not well defined on windows",
    )
    def test_parents_end_at_anchor(self):
        super().test_parents_end_at_anchor()


def test_multiple_backend_paths(local_testdir):
    path = "s3://bucket/"
    s3_path = UPath(path, anon=True)
    assert s3_path.joinpath("text.txt")._url.scheme == "s3"
    path = f"file://{local_testdir}"
    UPath(path)
    assert s3_path.joinpath("text1.txt")._url.scheme == "s3"


def test_constructor_accept_path(local_testdir):
    path = UPath(pathlib.Path(local_testdir))
    assert str(path) == str(pathlib.Path(local_testdir))


def test_constructor_accept_upath(local_testdir):
    path = UPath(UPath(local_testdir))
    assert str(path) == str(pathlib.Path(local_testdir))


def test_subclass(local_testdir):
    class MyPath(UPath):
        pass

    with pytest.raises(ValueError, match=r".*incompatible with"):
        MyPath(local_testdir)


@pytest.fixture(scope="function")
def upath_registry_snapshot():
    """Save and restore the upath registry state around a test."""
    from upath.registry import _registry

    # Save the current state of the registry's mutable mapping
    saved_m = _registry._m.maps[0].copy()
    try:
        yield
    finally:
        # Restore the registry state
        _registry._m.maps[0].clear()
        _registry._m.maps[0].update(saved_m)
        get_upath_class.cache_clear()


def test_subclass_registered(upath_registry_snapshot):
    class MyPath(UPath):
        pass

    register_implementation("memory", MyPath, clobber=True)
    path = MyPath("memory:///test_path")
    assert str(path) == "memory:///test_path"
    assert issubclass(MyPath, UPath)
    assert isinstance(path, MyPath)
    assert isinstance(path, pathlib_abc.ReadablePath)
    assert isinstance(path, pathlib_abc.WritablePath)
    assert not isinstance(path, pathlib.Path)


def test_subclass_with_gcs():
    path = UPath("gcs://bucket", anon=True)
    assert isinstance(path, UPath)
    assert isinstance(path, ReadablePath)
    assert isinstance(path, WritablePath)
    assert not isinstance(path, os.PathLike)
    assert not isinstance(path, pathlib.Path)


def test_instance_check(local_testdir):
    path = UPath(local_testdir)
    # test instance check passes
    assert isinstance(path, UPath)
    assert isinstance(path, ReadablePath)
    assert isinstance(path, WritablePath)
    assert isinstance(path, os.PathLike)
    assert isinstance(path, pathlib.Path)


def test_instance_check_local_uri(local_testdir):
    path = UPath(f"file://{local_testdir}")
    assert isinstance(path, UPath)
    assert isinstance(path, ReadablePath)
    assert isinstance(path, WritablePath)
    assert isinstance(path, os.PathLike)
    assert not isinstance(path, pathlib.Path)


@pytest.mark.xfail(reason="unsupported on universal_pathlib>0.1.4")
def test_new_method(local_testdir):
    path = UPath.__new__(pathlib.Path, local_testdir)
    assert str(path) == str(pathlib.Path(local_testdir))
    assert isinstance(path, pathlib.Path)
    assert isinstance(path, UPath)


PATHS = (
    ("path", "storage_options", "module", "object_type"),
    (
        ("/tmp/abc", {}, None, pathlib.Path),
        ("s3://bucket/folder", {"anon": True}, "s3fs", S3Path),
        ("gs://bucket/folder", {"token": "anon"}, "gcsfs", GCSPath),
    ),
)


@pytest.mark.parametrize(*PATHS)
def test_create_from_type(path, storage_options, module, object_type):
    """Test that derived paths use same fs instance."""
    if module:
        # skip if module cannot be imported
        pytest.importorskip(module)
    upath = UPath(path, **storage_options)
    # test expected object type
    assert isinstance(upath, object_type)
    cast = type(upath)
    parent = upath.parent
    # test derived object is same type
    assert isinstance(parent, cast)
    # test that created fs uses fsspec instance cache
    assert upath.fs is parent.fs
    new = cast(str(parent), **storage_options)
    # test that object cast is same type
    assert isinstance(new, cast)


def test_list_args():
    path_a = UPath("gcs://bucket", "folder")
    path_b = UPath("gcs://bucket") / "folder"

    assert str(path_a) == str(path_b)
    assert path_a.root == path_b.root
    assert path_a.drive == path_b.drive
    assert path_a.parts == path_b.parts
    assert path_a._url == path_b._url


def test_child_path():
    path_a = UPath("gcs://bucket/folder")
    path_b = UPath("gcs://bucket") / "folder"

    assert str(path_a) == str(path_b)
    assert path_a.root == path_b.root
    assert path_a.drive == path_b.drive
    assert path_a.parts == path_b.parts
    assert path_a._url == path_b._url


def test_pickling():
    path = UPath("gcs://bucket/folder", token="anon")
    pickled_path = pickle.dumps(path)
    recovered_path = pickle.loads(pickled_path)

    assert type(path) is type(recovered_path)
    assert str(path) == str(recovered_path)
    assert path.storage_options == recovered_path.storage_options


def test_pickling_child_path():
    path = UPath("gcs://bucket", token="anon") / "subfolder" / "subsubfolder"
    pickled_path = pickle.dumps(path)
    recovered_path = pickle.loads(pickled_path)

    assert type(path) is type(recovered_path)
    assert str(path) == str(recovered_path)
    assert path.drive == recovered_path.drive
    assert path.root == recovered_path.root
    assert path.parts == recovered_path.parts
    assert path.storage_options == recovered_path.storage_options


def test_copy_path():
    path = UPath("gcs://bucket/folder", token="anon")
    copy_path = UPath(path)

    assert type(path) is type(copy_path)
    assert str(path) == str(copy_path)
    assert path.drive == copy_path.drive
    assert path.root == copy_path.root
    assert path.parts == copy_path.parts
    assert path.storage_options == copy_path.storage_options


def test_copy_path_posix():
    path = UPath("/tmp/folder")
    copy_path = UPath(path)

    assert type(path) is type(copy_path)
    assert str(path) == str(copy_path)
    assert path.drive == copy_path.drive
    assert path.root == copy_path.root
    assert path.parts == copy_path.parts


def test_copy_path_append():
    path = UPath("/tmp/folder")
    copy_path = UPath(path, "folder2")

    assert type(path) is type(copy_path)
    assert str(path / "folder2") == str(copy_path)

    path = UPath("/tmp/folder")
    copy_path = UPath(path, "folder2/folder3")

    assert str(path / "folder2" / "folder3") == str(copy_path)

    path = UPath("/tmp/folder")
    copy_path = UPath(path, "folder2", "folder3")

    assert str(path / "folder2" / "folder3") == str(copy_path)


def test_compare_to_pathlib_path_ne():
    assert UPath("gcs://bucket/folder") != pathlib.Path("gcs://bucket/folder")
    assert pathlib.Path("gcs://bucket/folder") != UPath("gcs://bucket/folder")
    assert UPath("/bucket/folder") == pathlib.Path("/bucket/folder")
    assert pathlib.Path("/bucket/folder") == UPath("/bucket/folder")


def test_handle_fspath_args(tmp_path):
    f = tmp_path.joinpath("file.txt").as_posix()

    class X:
        def __str__(self):
            raise ValueError("should not be called")

        def __fspath__(self):
            return f

    assert UPath(X()).path == f


@pytest.mark.parametrize(
    "urlpath",
    [
        os.getcwd(),
        pathlib.Path.cwd().as_uri(),
        pytest.param(
            "mock:///abc",
            marks=pytest.mark.skipif(
                os.name == "nt",
                reason="_url not well defined for mock filesystem on windows",
            ),
        ),
    ],
)
def test_access_to_private_kwargs_and_url(urlpath):
    p0 = UPath(urlpath)
    assert not hasattr(p0, "_kwargs")

    # fixme: this should be deprecated...
    assert isinstance(p0._url, SplitResult)
    assert p0._url.scheme == "" or p0._url.scheme in p0.fs.protocol
    assert p0._url.path == p0.path

    p1 = p0 / "foo"
    assert isinstance(p1._url, SplitResult)
    assert p1._url.scheme == "" or p1._url.scheme in p1.fs.protocol
    assert p1._url.path == p1.path


def test_copy_path_append_kwargs():
    path = UPath("gcs://bucket/folder", anon=True)
    copy_path = UPath(path, anon=False)

    assert type(path) is type(copy_path)
    assert str(path) == str(copy_path)
    assert not copy_path.storage_options["anon"]
    assert path.storage_options["anon"]


def test_relative_to():
    assert "file.txt" == str(
        UPath("s3://test_bucket/file.txt").relative_to(UPath("s3://test_bucket"))
    )

    with pytest.raises(ValueError):
        UPath("s3://test_bucket/file.txt").relative_to(UPath("gcs://test_bucket"))

    with pytest.raises(ValueError):
        UPath("s3://test_bucket/file.txt", anon=True).relative_to(
            UPath("s3://test_bucket", anon=False)
        )


def test_uri_parsing():
    assert (
        str(UPath("http://www.example.com//a//b/")) == "http://www.example.com//a//b/"
    )


NORMALIZATIONS = (
    ("unnormalized", "normalized"),
    (
        # Normalization with and without an authority component
        ("memory:/a/b/..", "memory://a/"),
        ("memory:/a/b/.", "memory://a/b/"),
        ("memory:/a/b/../..", "memory://"),
        ("memory:/a/b/../../..", "memory://"),
        ("memory://a/b/.", "memory://a/b/"),
        ("memory://a/b/..", "memory://a/"),
        ("memory://a/b/../..", "memory://"),
        ("memory://a/b/../../..", "memory://"),
        ("memory:///a/b/.", "memory://a/b/"),
        ("memory:///a/b/..", "memory://a/"),
        ("memory:///a/b/../..", "memory://"),
        ("memory:///a/b/../../..", "memory://"),
    ),
)


@pytest.mark.parametrize(*NORMALIZATIONS)
def test_normalize(unnormalized, normalized):
    expected = UPath(normalized)
    pth = UPath(unnormalized)
    result = pth.resolve(strict=True)
    str_expected = str(expected)
    str_result = str(result)
    assert expected == result
    assert str_expected == str_result


@pytest.mark.parametrize(
    "uri,query_str",
    [
        ("s3://bucket/folder?versionId=1", "?versionId=1"),
        ("http://example.com/abc?p=2", "?p=2"),
    ],
)
def test_query_string(uri, query_str):
    p = UPath(uri)
    assert str(p).endswith(query_str)
    assert p.path.endswith(query_str)


PROTOCOL_MISMATCH = [
    ("/a", "s3://bucket/b"),
    ("s3://bucket/a", "gs://b/c"),
    ("gs://bucket/a", "memory://b/c"),
    ("memory://bucket/a", "s3://b/c"),
]


@pytest.mark.parametrize("base,join", PROTOCOL_MISMATCH)
def test_joinpath_on_protocol_mismatch(base, join):
    with pytest.raises(ValueError, match="can't combine incompatible UPath protocols"):
        UPath(base).joinpath(UPath(join))


@pytest.mark.parametrize("base,join", PROTOCOL_MISMATCH)
def test_truediv_on_protocol_mismatch(base, join):
    with pytest.raises(ValueError, match="can't combine incompatible UPath protocols"):
        UPath(base) / UPath(join)


@pytest.mark.parametrize("base,join", PROTOCOL_MISMATCH)
def test_joinuri_on_protocol_mismatch(base, join):
    assert UPath(base).joinuri(UPath(join)) == UPath(join)


def test_upath_expanduser():
    assert UPath("~").expanduser() == UPath(os.path.expanduser("~"))
    assert UPath("~") != UPath("~").expanduser()


def test_builtin_open_a_non_local_upath():
    p = UPath("memory://a")
    p.write_bytes(b"hello world")
    with pytest.raises(TypeError, match="expected str, bytes or os.PathLike object.*"):
        open(p, "rb")  # type: ignore


@pytest.mark.parametrize(
    "protocol",
    [
        pytest.param(None, id="empty protocol"),
        pytest.param("file", id="file protocol"),
    ],
)
def test_open_a_local_upath(tmp_path, protocol):
    p = tmp_path.joinpath("file.txt")
    p.write_bytes(b"hello world")
    u = UPath(p, protocol=protocol)
    with open(u, "rb") as f:
        assert f.read() == b"hello world"


@pytest.mark.parametrize(
    "uri,protocol",
    [
        # s3 compatible protocols
        ("s3://bucket/folder", "s3"),
        ("s3a://bucket/folder", "s3a"),
        ("bucket/folder", "s3"),
        # gcs compatible
        ("gs://bucket/folder", "gs"),
        ("gcs://bucket/folder", "gcs"),
        ("bucket/folder", "gs"),
        # azure compatible
        ("az://container/blob", "az"),
        ("abfs://container/blob", "abfs"),
        ("abfss://container/blob", "abfss"),
        ("adl://container/blob", "adl"),
        # memory
        ("memory://folder", "memory"),
        ("/folder", "memory"),
        # file/local
        ("file:/tmp/folder", "file"),
        ("/tmp/folder", "file"),
        ("file:/tmp/folder", "local"),
        ("/tmp/folder", "local"),
        ("/tmp/folder", ""),
        ("a/b/c", ""),
        # http/https
        ("http://example.com/path", "http"),
        ("https://example.com/path", "https"),
        # ftp
        ("ftp://example.com/path", "ftp"),
        # sftp/ssh
        ("sftp://example.com/path", "sftp"),
        ("ssh://example.com/path", "ssh"),
        # smb
        ("smb://server/share/path", "smb"),
        # hdfs
        ("hdfs://namenode/path", "hdfs"),
        # webdav - requires base_url, skip for now
        # github
        ("github://owner:repo@branch/path", "github"),
        # data
        ("data:text/plain;base64,SGVsbG8=", "data"),
        # huggingface
        ("hf://datasets/user/repo/path", "hf"),
    ],
)
def test_constructor_compatible_protocol_uri(uri, protocol):
    p = UPath(uri, protocol=protocol)
    assert p.protocol == protocol


# Protocol to sample URI mapping
_PROTOCOL_URIS = {
    "s3": "s3://bucket/folder",
    "gs": "gs://bucket/folder",
    "az": "az://container/blob",
    "memory": "memory://folder",
    "file": "file:/tmp/folder",
    "http": "http://example.com/path",
    "ftp": "ftp://example.com/path",
    "sftp": "sftp://example.com/path",
    "smb": "smb://server/share/path",
    "hdfs": "hdfs://namenode/path",
}

# Generate incompatible combinations: each protocol with URIs from other protocols
_INCOMPATIBLE_CASES = [
    (_PROTOCOL_URIS[uri_protocol], target_protocol)
    for target_protocol in _PROTOCOL_URIS
    for uri_protocol in _PROTOCOL_URIS
    if target_protocol != uri_protocol
]

# Also test explicit empty protocol with protocol-prefixed URIs
_INCOMPATIBLE_CASES.extend([(uri, "") for uri in _PROTOCOL_URIS.values()])


@pytest.mark.parametrize("uri,protocol", _INCOMPATIBLE_CASES)
def test_constructor_incompatible_protocol_uri(uri, protocol):
    with pytest.raises(TypeError, match=r".*incompatible with"):
        UPath(uri, protocol=protocol)


# Test subclass instantiation with incompatible URIs
# Use protocols that have registered implementations we can get via get_upath_class
_SUBCLASS_INCOMPATIBLE_CASES = [
    (_PROTOCOL_URIS[uri_protocol], target_protocol)
    for target_protocol in _PROTOCOL_URIS
    for uri_protocol in _PROTOCOL_URIS
    if target_protocol != uri_protocol
]


@pytest.mark.parametrize("uri,protocol", _SUBCLASS_INCOMPATIBLE_CASES)
def test_subclass_constructor_incompatible_protocol_uri(uri, protocol):
    cls = get_upath_class(protocol)
    with pytest.raises(TypeError, match=r".*incompatible with"):
        cls(uri)



from __future__ import annotations

import stat
import sys
import warnings
from typing import TYPE_CHECKING

from upath._stat import UPathStatResult
from upath.core import UnsupportedOperation
from upath.core import UPath
from upath.types import JoinablePathLike
from upath.types import StatResultType

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Literal

    if sys.version_info >= (3, 11):
        from typing import Self
        from typing import Unpack
    else:
        from typing_extensions import Self
        from typing_extensions import Unpack

    from upath._chain import FSSpecChainParser
    from upath.types.storage_options import TarStorageOptions


__all__ = ["TarPath"]


class TarPath(UPath):
    __slots__ = ()

    if TYPE_CHECKING:

        def __init__(
            self,
            *args: JoinablePathLike,
            protocol: Literal["zip"] | None = ...,
            chain_parser: FSSpecChainParser = ...,
            **storage_options: Unpack[TarStorageOptions],
        ) -> None: ...

    def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
        raise UnsupportedOperation

    def mkdir(
        self,
        mode: int = 0o777,
        parents: bool = False,
        exist_ok: bool = False,
    ) -> None:
        raise UnsupportedOperation

    def unlink(self, missing_ok: bool = False) -> None:
        raise UnsupportedOperation

    def write_bytes(self, data: bytes) -> int:
        raise UnsupportedOperation("DataPath does not support writing")

    def write_text(
        self,
        data: str,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> int:
        raise UnsupportedOperation("DataPath does not support writing")

    def stat(
        self,
        *,
        follow_symlinks: bool = True,
    ) -> StatResultType:
        if not follow_symlinks:
            warnings.warn(
                f"{type(self).__name__}.stat(follow_symlinks=False):"
                " is currently ignored.",
                UserWarning,
                stacklevel=2,
            )
        info = self.fs.info(self.path).copy()
        # convert mode
        if info["type"] == "directory":
            info["mode"] = stat.S_IFDIR
        elif info["type"] == "file":
            info["mode"] = stat.S_IFREG
        return UPathStatResult.from_info(info)

    def iterdir(self) -> Iterator[Self]:
        it = iter(super().iterdir())
        p0 = next(it)
        if p0.name != "":
            yield p0
        yield from it



import os
import shlex
import shutil
import subprocess
import sys
import threading
import time
import uuid
from pathlib import Path

import fsspec
import pytest
from fsspec import get_filesystem_class
from fsspec.implementations.local import LocalFileSystem
from fsspec.implementations.local import make_path_posix
from fsspec.implementations.smb import SMBFileSystem
from fsspec.registry import _registry
from fsspec.registry import register_implementation
from fsspec.utils import stringify_path
from packaging.version import Version

from .utils import posixify


class DummyTestFS(LocalFileSystem):
    protocol = "mock"
    root_marker = "/"

    @classmethod
    def _strip_protocol(cls, path):
        path = stringify_path(path)
        if path.startswith("mock://"):
            path = path[7:]
        elif path.startswith("mock:"):
            path = path[5:]
        return make_path_posix(path).rstrip("/") or cls.root_marker

    def unstrip_protocol(self, path):
        return f"mock://{self._strip_protocol(path)}"


@pytest.fixture(scope="session")
def clear_registry():
    register_implementation("mock", DummyTestFS)
    try:
        yield
    finally:
        _registry.clear()


@pytest.fixture(scope="function")
def windows_working_directory_drive_sync(monkeypatch, tmp_path, tmp_path_factory):
    cwd_old = os.getcwd()
    drive_cwd = os.path.splitdrive(cwd_old)[0]
    drive_tmp = os.path.splitdrive(tmp_path)[0]
    if drive_tmp != drive_cwd:
        cwd_new = tmp_path_factory.mktemp("cwd_on_tmp_drive")
        os.chdir(cwd_new)
        try:
            yield
        finally:
            os.chdir(cwd_old)
    else:
        yield


@pytest.fixture(scope="function")
def clear_fsspec_memory_cache():
    fs_cls = get_filesystem_class("memory")
    pseudo_dirs = fs_cls.pseudo_dirs.copy()
    store = fs_cls.store.copy()
    try:
        yield
    finally:
        fs_cls.pseudo_dirs = pseudo_dirs
        fs_cls.store = store


@pytest.fixture(scope="function")
def local_testdir(tmp_path, clear_registry):
    folder1 = tmp_path.joinpath("folder1")
    folder1.mkdir()
    folder1_files = ["file1.txt", "file2.txt"]
    for f in folder1_files:
        p = folder1.joinpath(f)
        p.touch()
        p.write_text(f)

    file1 = tmp_path.joinpath("file1.txt")
    file1.touch()
    file1.write_text("hello world")
    file2 = tmp_path.joinpath("file2.txt")
    file2.touch()
    file2.write_bytes(b"hello world")
    if sys.platform.startswith("win"):
        yield str(tmp_path).replace("\\", "/")
    else:
        yield str(tmp_path)


@pytest.fixture()
def pathlib_base(local_testdir):
    return Path(local_testdir)


@pytest.fixture(scope="session")
def htcluster():
    try:
        proc = subprocess.Popen(
            shlex.split("htcluster startup"),
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )
    except FileNotFoundError as err:
        if err.errno == 2 and "htcluster" == err.filename:
            pytest.skip("htcluster not installed")
        raise

    time.sleep(30)
    try:
        yield
    finally:
        proc.terminate()
        proc.wait()
        proc1 = subprocess.Popen(
            shlex.split("htcluster shutdown"),
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )
        proc1.terminate()
        proc1.wait()
        time.sleep(10)


@pytest.fixture()
def hdfs(htcluster, tmp_path, local_testdir):
    pyarrow = pytest.importorskip("pyarrow")
    host, user, port = "0.0.0.0", "hdfs", 9000
    hdfs = pyarrow.hdfs.connect(host="0.0.0.0", port=9000, user=user)
    hdfs.mkdir(str(tmp_path).encode("utf8"), create_parents=True)
    for x in Path(local_testdir).glob("**/*"):
        if x.is_file():
            text = x.read_text().encode("utf8")
            if not hdfs.exists(str(x.parent)):
                hdfs.mkdir(str(x.parent), create_parents=True)
            with hdfs.open(str(x), "wb") as f:
                f.write(text)
        else:
            hdfs.mkdir(str(x))
    hdfs.close()
    yield host, user, port


@pytest.fixture(scope="session")
def s3_server():
    # writable local S3 system
    if "BOTO_CONFIG" not in os.environ:  # pragma: no cover
        os.environ["BOTO_CONFIG"] = "/dev/null"
    if "AWS_ACCESS_KEY_ID" not in os.environ:  # pragma: no cover
        os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    if "AWS_SECRET_ACCESS_KEY" not in os.environ:  # pragma: no cover
        os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    if "AWS_SECURITY_TOKEN" not in os.environ:  # pragma: no cover
        os.environ["AWS_SECURITY_TOKEN"] = "testing"
    if "AWS_SESSION_TOKEN" not in os.environ:  # pragma: no cover
        os.environ["AWS_SESSION_TOKEN"] = "testing"
    if "AWS_DEFAULT_REGION" not in os.environ:  # pragma: no cover
        os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
    requests = pytest.importorskip("requests")

    pytest.importorskip("moto")

    port = 5555
    endpoint_uri = f"http://127.0.0.1:{port}/"
    proc = subprocess.Popen(
        [sys.executable, *shlex.split(f"-m moto.server -p {port}")],
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
    )
    try:
        timeout = 5
        while timeout > 0:
            try:
                r = requests.get(endpoint_uri, timeout=10)
                if r.ok:
                    break
            except Exception:  # pragma: no cover
                pass
            timeout -= 0.1  # pragma: no cover
            time.sleep(0.1)  # pragma: no cover
        anon = False
        s3so = {
            "client_kwargs": {"endpoint_url": endpoint_uri},
            "use_listings_cache": True,
        }
        yield anon, s3so
    finally:
        proc.terminate()
        proc.wait()


@pytest.fixture
def s3_fixture(s3_server, local_testdir):
    pytest.importorskip("s3fs")
    anon, s3so = s3_server
    s3 = fsspec.filesystem("s3", anon=False, **s3so)
    bucket_name = "test_bucket"
    if s3.exists(bucket_name):
        for dir, _, keys in s3.walk(bucket_name):
            for key in keys:
                s3.rm(f"{dir}/{key}")
    else:
        s3.mkdir(bucket_name)
    for x in Path(local_testdir).glob("**/*"):
        target_path = f"{bucket_name}/{posixify(x.relative_to(local_testdir))}"
        if x.is_file():
            s3.upload(str(x), target_path)
    s3.invalidate_cache()
    yield f"s3://{bucket_name}", anon, s3so


def stop_docker(container):
    cmd = shlex.split('docker ps -a -q --filter "name=%s"' % container)
    cid = subprocess.check_output(cmd).strip().decode()
    if cid:
        subprocess.call(["docker", "rm", "-f", "-v", cid])


@pytest.fixture(scope="session")
def docker_gcs():
    if "STORAGE_EMULATOR_HOST" in os.environ:
        # assume using real API or otherwise have a server already set up
        yield os.environ["STORAGE_EMULATOR_HOST"]
        return

    requests = pytest.importorskip("requests")
    if shutil.which("docker") is None:
        pytest.skip("docker not installed")

    container = "gcsfs_test"
    cmd = " ".join(
        [
            "docker",
            "run",
            "-d",
            "-p 4443:4443",
            "--name gcsfs_test",
            "fsouza/fake-gcs-server:latest",
            "-scheme http",
            "-public-host http://localhost:4443",
            "-external-url http://localhost:4443",
            "-backend memory",
        ]
    )
    stop_docker(container)
    subprocess.check_output(shlex.split(cmd))
    url = "http://0.0.0.0:4443"
    timeout = 10
    while True:
        try:
            r = requests.get(url + "/storage/v1/b", timeout=10)
            if r.ok:
                yield url
                break
        except Exception as e:  # noqa: E722
            timeout -= 1
            if timeout < 0:
                raise SystemError from e
            time.sleep(1)
    stop_docker(container)


@pytest.fixture
def gcs_fixture(docker_gcs, local_testdir):
    pytest.importorskip("gcsfs")
    gcs = fsspec.filesystem("gcs", endpoint_url=docker_gcs, token="anon")
    bucket_name = "test_bucket"
    if gcs.exists(bucket_name):
        for dir, _, keys in gcs.walk(bucket_name):
            for key in keys:
                gcs.rm(f"{dir}/{key}")
    else:
        gcs.mkdir(bucket_name)
    for x in Path(local_testdir).glob("**/*"):
        target_path = f"{bucket_name}/{posixify(x.relative_to(local_testdir))}"
        if x.is_file():
            gcs.upload(str(x), target_path)
    gcs.invalidate_cache()
    yield f"gs://{bucket_name}", docker_gcs


@pytest.fixture(scope="session")
def http_server(tmp_path_factory):
    http_tempdir = tmp_path_factory.mktemp("http")

    requests = pytest.importorskip("requests")
    pytest.importorskip("http.server")
    proc = subprocess.Popen(
        shlex.split(f"{sys.executable} -m http.server --directory {http_tempdir} 18080")
    )
    try:
        url = "http://127.0.0.1:18080/folder"
        path = Path(http_tempdir) / "folder"
        path.mkdir()
        timeout = 10
        while True:
            try:
                r = requests.get(url, timeout=10)
                if r.ok:
                    yield path, url
                    break
            except Exception as e:  # noqa: E722
                timeout -= 1
                if timeout < 0:
                    raise SystemError from e
                time.sleep(1)
    finally:
        proc.terminate()
        proc.wait()


@pytest.fixture
def http_fixture(local_testdir, http_server):
    http_path, http_url = http_server
    shutil.rmtree(http_path)
    shutil.copytree(local_testdir, http_path)
    yield http_url


@pytest.fixture(scope="session")
def webdav_server(tmp_path_factory):
    try:
        from cheroot import wsgi
        from wsgidav.wsgidav_app import WsgiDAVApp
    except ImportError as err:
        pytest.skip(str(err))

    webdav_tmp_dir = str(tmp_path_factory.mktemp("webdav"))

    host = "127.0.0.1"
    port = 8090
    app = WsgiDAVApp(
        {
            "host": host,
            "port": port,
            "provider_mapping": {"/": webdav_tmp_dir},
            "simple_dc": {"user_mapping": {"*": {"USER": {"password": "PASSWORD"}}}},
        }
    )
    srvr = wsgi.Server(bind_addr=(host, port), wsgi_app=app)
    srvr.prepare()
    thread = threading.Thread(target=srvr.serve, daemon=True)
    thread.start()

    try:
        yield f"webdav+http://{host}:{port}", app
    finally:
        srvr.stop()


@pytest.fixture
def webdav_fixture(local_testdir, webdav_server):
    webdav_url, app = webdav_server
    # switch to new test directory
    fs_provider = app.provider_map["/"]
    fs_provider.root_folder_path = os.path.abspath(local_testdir)
    try:
        yield webdav_url
    finally:
        # clear locks if any are held
        fs_provider.lock_manager.storage.clear()


AZURITE_PORT = int(os.environ.get("UPATH_AZURITE_PORT", "10000"))


@pytest.fixture(scope="session")
def azurite_credentials():
    url = f"http://localhost:{AZURITE_PORT}"
    account_name = "devstoreaccount1"
    key = "Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw=="  # noqa: E501
    endpoint = f"{url}/{account_name}"
    connection_string = f"DefaultEndpointsProtocol=http;AccountName={account_name};AccountKey={key};BlobEndpoint={endpoint};"  # noqa

    yield account_name, connection_string


@pytest.fixture(scope="session")
def docker_azurite(azurite_credentials):
    requests = pytest.importorskip("requests")

    if shutil.which("docker") is None:
        pytest.skip("docker not installed")

    image = "mcr.microsoft.com/azure-storage/azurite"
    container_name = "azure_test"
    cmd = (
        f"docker run --rm -d -p {AZURITE_PORT}:10000 --name {container_name} {image}:latest"  # noqa: E501
        " azurite-blob --loose --blobHost 0.0.0.0 --skipApiVersionCheck"  # noqa: E501
    )
    url = f"http://localhost:{AZURITE_PORT}"

    stop_docker(container_name)
    subprocess.run(shlex.split(cmd), check=True)

    retries = 10
    while True:
        try:
            # wait until the container is up, even a 400 status code is ok
            r = requests.get(url, timeout=10)
            if (
                r.status_code == 400
                and "Server" in r.headers
                and "Azurite" in r.headers["Server"]
            ):
                yield url
                break
        except Exception as e:  # noqa: E722
            retries -= 1
            if retries < 0:
                raise SystemError from e
            time.sleep(1)

    stop_docker(container_name)


@pytest.fixture(scope="session")
def azure_container(azurite_credentials, docker_azurite):
    azure_storage = pytest.importorskip("azure.storage.blob")
    account_name, connection_string = azurite_credentials
    client = azure_storage.BlobServiceClient.from_connection_string(
        conn_str=connection_string
    )
    container_name = str(uuid.uuid4())
    client.create_container(container_name)

    try:
        yield container_name
    finally:
        client.delete_container(container_name)


@pytest.fixture(scope="function")
def azure_fixture(azurite_credentials, azure_container):
    azure_storage = pytest.importorskip("azure.storage.blob")
    account_name, connection_string = azurite_credentials
    client = azure_storage.BlobServiceClient.from_connection_string(
        conn_str=connection_string
    ).get_container_client(azure_container)

    try:
        yield f"az://{azure_container}"
    finally:
        for blob in client.list_blobs():
            client.delete_blob(blob["name"])


@pytest.fixture(scope="module")
def smb_container():
    try:
        pchk = ["docker", "run", "--name", "fsspec_test_smb", "hello-world"]
        subprocess.check_call(pchk)
        stop_docker("fsspec_test_smb")
    except (subprocess.CalledProcessError, FileNotFoundError):
        pytest.skip("docker run not available")

    # requires docker
    container = "fsspec_smb"
    stop_docker(container)
    cfg = "-p -u 'testuser;testpass' -s 'home;/share;no;no;no;testuser'"
    port = int(os.environ.get("UPATH_TESTS_SMB_PORT", "445"))
    img = f"docker run --name {container} --detach -p 139:139 -p {port}:445 dperson/samba"  # noqa: E231 E501
    cmd = f"{img} {cfg}"
    try:
        subprocess.check_output(shlex.split(cmd)).strip().decode()
        time.sleep(2)
        yield {
            "host": "localhost",
            "port": port,
            "username": "testuser",
            "password": "testpass",
            "register_session_retries": 100,  # max ~= 10 seconds
        }
    finally:
        import smbclient  # pylint: disable=import-outside-toplevel

        smbclient.reset_connection_cache()
        stop_docker(container)


@pytest.fixture
def smb_url(smb_container):
    smb_url = "smb://{username}:{password}@{host}:{port}/home/"
    smb_url = smb_url.format(**smb_container)
    return smb_url


@pytest.fixture
def smb_fixture(local_testdir, smb_url, smb_container):
    smb = SMBFileSystem(
        host=smb_container["host"],
        port=smb_container["port"],
        username=smb_container["username"],
        password=smb_container["password"],
    )
    url = smb_url + "testdir/"
    smb.put(local_testdir, "/home/testdir", recursive=True)
    yield url
    smb.delete("/home/testdir", recursive=True)


@pytest.fixture(scope="module")
def ssh_container():
    if shutil.which("docker") is None:
        pytest.skip("docker not installed")

    name = "fsspec_test_ssh"
    stop_docker(name)
    cmd = (
        "docker run"
        " -d"
        f" --name {name}"
        " -e USER_NAME=user"
        " -e PASSWORD_ACCESS=true"
        " -e USER_PASSWORD=pass"
        " -p 2222:2222"
        " linuxserver/openssh-server:latest"
    )
    try:
        subprocess.run(shlex.split(cmd))
        yield {
            "host": "localhost",
            "port": 2222,
            "username": "user",
            "password": "pass",
        }
    finally:
        stop_docker(name)


@pytest.fixture
def ssh_fixture(ssh_container, local_testdir, monkeypatch):
    paramiko = pytest.importorskip("paramiko", reason="sftp tests require paramiko")

    cls = fsspec.get_filesystem_class("ssh")
    if cls.put != fsspec.AbstractFileSystem.put:
        monkeypatch.setattr(cls, "put", fsspec.AbstractFileSystem.put)
    if Version(fsspec.__version__) < Version("2022.10.0"):
        from fsspec.callbacks import _DEFAULT_CALLBACK

        monkeypatch.setattr(_DEFAULT_CALLBACK, "relative_update", lambda *args: None)

    for _ in range(100):
        try:
            fs = fsspec.filesystem(
                "ssh",
                host=ssh_container["host"],
                port=ssh_container["port"],
                username=ssh_container["username"],
                password=ssh_container["password"],
                timeout=10.0,
                banner_timeout=30.0,
                skip_instance_cache=True,
            )
        except (
            paramiko.ssh_exception.NoValidConnectionsError,
            paramiko.ssh_exception.SSHException,
        ):
            time.sleep(0.1)
            continue
        break
    else:
        raise RuntimeError("issue with openssh-container startup")

    fs.put(local_testdir, "/app/testdir", recursive=True)
    try:
        yield "ssh://{username}:{password}@{host}:{port}/app/testdir/".format(
            **ssh_container
        )
    finally:
        fs.delete("/app/testdir", recursive=True)


@pytest.fixture
def hf_test_repo():
    # "__username__" is an invalid username so we can use it for tests
    return "__username__/test_repo"


@pytest.fixture
def mock_hf_api(pathlib_base, monkeypatch, hf_test_repo):  # noqa: C901
    huggingface_hub = pytest.importorskip(
        "huggingface_hub", reason="hf tests require huggingface_hub"
    )
    hf_file_system = pytest.importorskip(
        "huggingface_hub.hf_file_system", reason="hf tests require huggingface_hub"
    )
    httpx = pytest.importorskip("httpx")

    class MockedHfApi(huggingface_hub.HfApi):

        def repo_info(self, repo_id, *args, repo_type=None, **kwargs):
            if repo_id != hf_test_repo:
                raise huggingface_hub.errors.RepositoryNotFoundError(
                    repo_id,
                    response=httpx.Response(404, request=...),
                )
            elif repo_type is None or repo_type == "model":
                return huggingface_hub.hf_api.ModelInfo(id=repo_id)
            elif repo_type == "dataset":
                return huggingface_hub.hf_api.DatasetInfo(id=repo_id)
            elif repo_type == "space":
                return huggingface_hub.hf_api.SpaceInfo(id=repo_id)
            else:
                raise ValueError("Unsupported repo type.")

        def get_paths_info(self, repo_id, paths, *args, **kwargs):
            if repo_id != hf_test_repo:
                raise huggingface_hub.errors.RepositoryNotFoundError(
                    repo_id,
                    response=httpx.Response(404, request=...),
                )
            paths_info = []
            for path in paths:
                if path:
                    path = pathlib_base / path
                    if path.is_file():
                        paths_info.append(
                            huggingface_hub.hf_api.RepoFile(
                                path=path.relative_to(pathlib_base).as_posix(),
                                blob_id="blob_id",
                                size=path.stat().st_size,
                            )
                        )
                    elif path.is_dir():
                        paths_info.append(
                            huggingface_hub.hf_api.RepoFolder(
                                path=path.relative_to(pathlib_base).as_posix(),
                                tree_id="tree_id",
                            )
                        )
            return paths_info

        def list_repo_tree(
            self, repo_id, path_in_repo, *args, recursive=False, **kwargs
        ):
            if repo_id != hf_test_repo:
                raise huggingface_hub.errors.RepositoryNotFoundError(
                    repo_id, response=httpx.Response(404, request=...)
                )
            pathlib_dir = pathlib_base / path_in_repo if path_in_repo else pathlib_base
            for path in pathlib_dir.rglob("*") if recursive else pathlib_dir.glob("*"):
                if path.is_file():
                    yield huggingface_hub.hf_api.RepoFile(
                        path=path.relative_to(pathlib_base).as_posix(),
                        oid="oid",
                        size=path.stat().st_size,
                    )
                else:
                    yield huggingface_hub.hf_api.RepoFolder(
                        path=path.relative_to(pathlib_base).as_posix(),
                        oid="oid",
                    )

    hf_file_system.HfFileSystem.clear_instance_cache()
    monkeypatch.setattr(hf_file_system, "HfApi", MockedHfApi)


@pytest.fixture
def mock_hf_filesystem_open(pathlib_base, monkeypatch):
    hf_file_system = pytest.importorskip(
        "huggingface_hub.hf_file_system", reason="hf tests require huggingface_hub"
    )

    def mocked_open(fs, path, mode="rb", *args, **kwargs):
        resolved_path = fs.resolve_path(path)
        return (pathlib_base / resolved_path.path_in_repo).open(mode)

    monkeypatch.setattr(hf_file_system.HfFileSystem, "_open", mocked_open)


@pytest.fixture
def hf_fixture_with_readonly_mocked_hf_api(
    hf_test_repo, mock_hf_api, mock_hf_filesystem_open
):
    return "hf://" + hf_test_repo


@pytest.fixture(scope="module")
def ftp_server_process(tmp_path_factory):
    """Fixture providing a writable FTP filesystem."""
    pytest.importorskip("pyftpdlib")

    tmp_path = tmp_path_factory.mktemp("ftp-server")

    P = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "pyftpdlib",
            "-d",
            str(tmp_path),
            "-u",
            "user",
            "-P",
            "pass",
            "-w",
        ]
    )
    try:
        time.sleep(1)
        yield str(tmp_path), {
            "host": "localhost",
            "port": 2121,
            "username": "user",
            "password": "pass",
        }
    finally:
        P.terminate()
        P.wait()
        try:
            shutil.rmtree(tmp_path)
        except Exception:
            pass


@pytest.fixture(scope="function")
def ftp_server(ftp_server_process):
    """Fixture providing a writable FTP filesystem."""
    tmp_path, storage_options = ftp_server_process

    try:
        yield storage_options
    finally:
        for filename in os.listdir(tmp_path):
            file_path = os.path.join(tmp_path, filename)
            if os.path.isdir(file_path):
                del_func = shutil.rmtree
            else:
                del_func = os.unlink
            try:
                del_func(file_path)
            except Exception:
                pass



import pytest

try:
    from pydantic import BaseConfig
    from pydantic_settings import BaseSettings
except ImportError:
    BaseConfig = BaseSettings = None
    pytestmark = pytest.mark.skip(reason="requires pydantic")

from upath.core import UPath


def test_pydantic_settings_local_upath():
    class MySettings(BaseSettings):
        example_path: UPath = UPath(__file__)

    assert isinstance(MySettings().example_path, UPath)



import pytest

from upath import UPath
from upath.implementations.sftp import SFTPPath

from ..cases import BaseTests
from ..utils import OverrideMeta
from ..utils import overrides_base
from ..utils import skip_on_windows


@skip_on_windows
class TestUPathSFTP(BaseTests, metaclass=OverrideMeta):

    @pytest.fixture(autouse=True)
    def path(self, ssh_fixture):
        self.path = UPath(ssh_fixture)

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, SFTPPath)


@pytest.mark.parametrize(
    "args,parts",
    [
        (("sftp://user@host",), ("/",)),
        (("sftp://user@host/",), ("/",)),
        (("sftp://user@host", ""), ("/",)),
        (("sftp://user@host/", ""), ("/",)),
        (("sftp://user@host", "/"), ("/",)),
        (("sftp://user@host/", "/"), ("/",)),
        (("sftp://user@host/abc",), ("/", "abc")),
        (("sftp://user@host", "abc"), ("/", "abc")),
        (("sftp://user@host", "/abc"), ("/", "abc")),
        (("sftp://user@host/", "/abc"), ("/", "abc")),
    ],
)
def test_join_produces_correct_parts(args, parts):
    pth = UPath(*args)
    assert pth.storage_options == {"host": "host", "username": "user"}
    assert pth.parts == parts
